"""
    Base Graph API impl
"""
import hashlib
import hmac
import logging
import re
import time
from urllib.parse import parse_qsl, urlparse
from typing import Dict, List, Optional, Tuple

import requests
from requests import Response
from requests_oauthlib.oauth2_session import OAuth2Session
from requests_oauthlib.compliance_fixes.facebook import facebook_compliance_fix

from pyfacebook import RateLimit, PercentSecond, FacebookError, LibraryError

logger = logging.getLogger(__name__)


class GraphAPI:
    VALID_API_VERSIONS = [
        "v7.0",
        "v8.0",
        "v9.0",
        "v10.0",
        "v11.0",
        "v12.0",
        "v13.0",
        "v14.0",
    ]
    GRAPH_URL = "https://graph.facebook.com/"
    AUTHORIZATION_URL = "https://www.facebook.com/dialog/oauth"
    EXCHANGE_ACCESS_TOKEN_URL = "https://graph.facebook.com/oauth/access_token"
    DEFAULT_REDIRECT_URI = "https://localhost/"
    DEFAULT_SCOPE = ["public_profile"]
    STATE = "PyFacebook"

    def __init__(
        self,
        app_id: Optional[str] = None,
        app_secret: Optional[str] = None,
        access_token: Optional[str] = None,
        application_only_auth: bool = False,
        oauth_flow: bool = False,
        version: Optional[str] = None,
        sleep_on_rate_limit: bool = True,
        sleep_seconds_mapping: Optional[Dict[int, int]] = None,
        base_url: Optional[str] = None,
        timeout: Optional[int] = None,
        proxies: Optional[dict] = None,
        instagram_business_id: Optional[str] = None,
        authorization_url: Optional[str] = None,
        access_token_url: Optional[str] = None,
    ):
        self.app_id = app_id
        self.app_secret = app_secret
        self.access_token = access_token

        self.session = requests.Session()
        self.__timeout = timeout
        self.proxies = proxies
        self.sleep_on_rate_limit = sleep_on_rate_limit
        self.sleep_seconds_mapping = self._build_sleep_seconds_resource(
            sleep_seconds_mapping=sleep_seconds_mapping
        )
        self.rate_limit = RateLimit()
        self.instagram_business_id = instagram_business_id

        # Override url for send request
        self.base_url = base_url if base_url else self.GRAPH_URL
        self.authorization_url = (
            authorization_url if authorization_url else self.AUTHORIZATION_URL
        )
        self.access_token_url = (
            access_token_url if access_token_url else self.EXCHANGE_ACCESS_TOKEN_URL
        )

        # version check
        if version is None:
            # default version is last new.
            self.version = self.VALID_API_VERSIONS[-1]
        else:
            version = str(version)
            if not version.startswith("v"):
                version = "v" + version
            version_regex = re.compile(r"^v\d*.\d{1,2}$")
            match = version_regex.search(str(version))
            if match is not None:
                self.version = version
            else:
                raise LibraryError(
                    {
                        "message": f"Invalid version {version}. You can provide with like: 14.0 or v14.0"
                    }
                )

        # Token
        if access_token:
            self.access_token = access_token
        elif application_only_auth and all([self.app_id, self.app_secret]):
            data = self.get_app_token()
            self.access_token = data["access_token"]
        elif oauth_flow and all([self.app_id, self.app_secret]):
            pass
        else:
            raise LibraryError({"message": "Need access token"})

    @staticmethod
    def _build_sleep_seconds_resource(
        sleep_seconds_mapping: Optional[Dict[int, int]]
    ) -> Optional[List[PercentSecond]]:
        """
        Sort and convert data
        :param sleep_seconds_mapping: mapping for sleep.
        :return:
        """
        if sleep_seconds_mapping is None:
            return None
        mapping_list = [
            PercentSecond(percent=p, seconds=s)
            for p, s in sleep_seconds_mapping.items()
        ]
        return sorted(mapping_list, key=lambda ps: ps.percent)

    @staticmethod
    def _generate_secret_proof(
        access_token: str, secret: Optional[str] = None
    ) -> Optional[str]:
        """
        :param access_token:
        :param secret: App secret
        :return:
        """
        if secret is None:
            logger.debug(
                "Calls from a server can be better secured by adding a parameter called appsecret_proof. "
                "And need your app secret."
            )
            return None
        return hmac.new(
            secret.encode("utf-8"),
            msg=access_token.encode("utf-8"),
            digestmod=hashlib.sha256,
        ).hexdigest()

    def _append_token(self, args: Optional[dict]) -> dict:
        """
        Append access token and secret_proof parameter of parameters.
        :param args: Original parameters.
        :return: New parameters.
        """
        args = {} if args is None else args
        if "access_token" not in args:
            args["access_token"] = self.access_token
        # Begin with v5.0, appsecret_proof parameter can improve requests secure.
        # Refer: https://developers.facebook.com/docs/graph-api/securing-requests/
        secret_proof = self._generate_secret_proof(
            args["access_token"], self.app_secret
        )
        args["appsecret_proof"] = secret_proof
        return args

    def _request(
        self,
        url: str,
        args: Optional[dict] = None,
        post_args: Optional[dict] = None,
        verb: str = "GET",
        auth_need: bool = True,
        **kwargs,
    ) -> Response:
        """
        :param url: Resource url for Graph.
        :param args: Query parameters.
        :param post_args: Form parameters.
        :param verb: HTTP method
        :param auth_need: Whether need access token.
        :param kwargs: Additional parameters.
        :return:
        """
        if auth_need:
            if verb == "GET" or verb == "DELETE":
                args = self._append_token(args=args)
            elif verb == "POST":
                post_args = self._append_token(args=post_args)

        if not url.startswith("http"):
            url = self.base_url + url

        try:
            response = self.session.request(
                method=verb,
                url=url,
                timeout=self.__timeout,
                params=args,
                data=post_args,
                proxies=self.proxies,
                **kwargs,
            )
        except requests.HTTPError as ex:
            raise LibraryError({"message": ex.args})

        # check headers
        headers = response.headers
        self.rate_limit.set_limit(headers)
        if self.sleep_on_rate_limit:
            sleep_seconds = self.rate_limit.get_sleep_seconds(
                sleep_data=self.sleep_seconds_mapping
            )
            time.sleep(sleep_seconds)
        return response

    def _parse_response(self, response: Response) -> dict:
        """
        :param response: Response from graph api.
        :return: json data
        """
        content_type = response.headers["Content-Type"]
        if "json" in content_type:
            data = response.json()
            self._check_graph_error(data=data)
            return data
        elif "image/" in content_type:
            data = {
                "data": response.content,
                "content-type": content_type,
                "url": response.url,
            }
            return data
        else:
            raise LibraryError({"message": "Wrong response, not json or image"})

    @staticmethod
    def _check_graph_error(data: dict):
        """
        :param data: Data from response
        """
        if "error" in data:
            raise FacebookError(data)

    def get(self, path, args):
        """
        Send GET request.

        :param path: path for resource.
        :param args: args for request.
        :return: Response data
        """
        resp = self._request(
            url=f"{self.version}/{path}",
            args=args,
        )
        data = self._parse_response(resp)
        return data

    def get_object(self, object_id: str, fields: str = "", **kwargs) -> dict:
        """
        Get object information by object id.

        :param object_id: ID for object(user,page,event...).
        :param fields: Comma-separated string for object fields which you want.
        :param kwargs: Additional parameters for object.
        :return: Response data
        """
        args = {"fields": fields}
        if kwargs:
            args.update(kwargs)

        resp = self._request(
            url=f"{self.version}/{object_id}",
            args=args,
        )
        data = self._parse_response(resp)
        return data

    def get_objects(self, ids: str, fields: str = "", **kwargs) -> dict:
        """
        Get objects information by multi object ids.

        :param ids: Comma-separated string for object ids which you want.
        :param fields: Comma-separated string for object fields which you want.
        :param kwargs: Additional parameters for object.
        :return: Response data
        """
        args = {"ids": ids, "fields": fields}
        if kwargs:
            args.update(kwargs)

        resp = self._request(url=f"{self.version}", args=args)
        data = self._parse_response(resp)
        return data

    def get_connection(
        self,
        object_id: str,
        connection: str,
        **kwargs,
    ) -> dict:
        """
        Get connections objects for object by id. Like get page medias by page id.

        :param object_id: ID for object(user,page,event...).
        :param connection: Connection name for object, Like(posts,comments...).
        :param kwargs: Additional parameters for different connections.
        :return: Response data
        """
        resp = self._request(
            url=f"{self.version}/{object_id}/{connection}", args=kwargs
        )
        data = self._parse_response(resp)
        return data

    def get_full_connections(
        self,
        object_id: str,
        connection: str,
        count: Optional[int] = 10,
        limit: Optional[int] = None,
        **kwargs,
    ) -> dict:
        """
        Get connections objects for object by id. Like get page medias by page id.

        :param object_id: ID for object(user,page,event...).
        :param connection: Connection name for object, Like(posts,comments...).
        :param count: The count will retrieve objects. Default is None will get all data.
        :param limit: Each request retrieve objects count.
            For most connections should no more than 100. Default is None will use api default limit.
        :param kwargs: Additional parameters for different connections.
        :return: Combined Response data
        """

        data, data_set, paging = {}, [], None
        while True:
            # sometimes may not return limit.
            if limit is not None:
                kwargs["limit"] = limit

            data = self.get_connection(
                object_id=object_id,
                connection=connection,
                **kwargs,
            )
            # Append this request data
            data_set.extend(data["data"])
            if count is not None and len(data_set) > count:
                data_set = data_set[:count]
                break

            # check next pagination
            paging, _next = data.get("paging"), None
            if paging is not None:
                _next = paging.get("next")
            if not _next:
                break
            # parse next url args as new args
            kwargs = dict(parse_qsl(urlparse(_next).query))

        # Replace the data list in data.
        data["data"] = data_set
        return data

    def discovery_user_media(
        self,
        username: str,
        fields: str = "",
        count: Optional[int] = 10,
        limit: Optional[int] = 10,
        since: Optional[str] = None,
        until: Optional[str] = None,
        **kwargs,
    ) -> dict:
        """
        Discovery other business account media.
        :param username: Username for the instagram account.
        :param fields: Comma-separated string for object fields which you want.
        :param count: The count will retrieve objects. Default is None will get all data.
        :param limit: Each request retrieve objects count.
            For most connections should no more than 100. Default is None will use api default limit.
        :param since: A Unix timestamp or strtotime data value that points to the start of data.
        :param until: A Unix timestamp or strtotime data value that points to the end of data.
        :return: Combined Response data
        """
        limit = f".limit({limit})" if limit is not None else ""
        since = f".since({since})" if since is not None else ""
        until = f".until({until})" if until is not None else ""
        after = kwargs.get("after", "")

        base_query = "business_discovery.username({username}){{media{after}{limit}{since}{until}{{{fields}}}}}"
        data, media_set, paging = {}, [], None
        while True:
            # next page for result
            after = f".after({after})" if after else ""

            fds = base_query.format(
                username=username,
                fields=fields,
                after=after,
                limit=limit,
                since=since,
                until=until,
            )
            args = {"fields": fds}

            data = self.get(
                path=self.instagram_business_id,
                args=args,
            )
            data = data.get("business_discovery", {}).get("media", {})
            # Append this request data
            if data:
                media_set.extend(data.get("data", []))
            if count is not None and len(media_set) > count:
                media_set = media_set[:count]
                break

            # check next pagination
            paging, after = data.get("paging"), None
            if paging is not None:
                after = paging.get("cursors", {}).get("after")
            if not after:
                break

        # Replace the data list in data.
        data["data"] = media_set
        return data

    def post_object(
        self,
        object_id: str,
        connection: Optional[str] = None,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        **kwargs,
    ) -> dict:
        """
        Create or update data for a facebook object or it's edge.

        :param object_id: ID for the facebook object(page,user.. and so on).
        :param connection: Edge for the object.
        :param params: Parameters for url path.
        :param data: Parameters for Form data.
        :param kwargs: Additional parameters.
        :return: Response data.
        """
        path = f"{self.version}/{object_id}"
        if connection:
            path += f"/{connection}"

        resp = self._request(
            url=path,
            args=params,
            post_args=data,
            verb="POST",
            **kwargs,
        )
        data = self._parse_response(resp)
        return data

    def delete_object(
        self,
        object_id: str,
        **kwargs,
    ) -> dict:
        """
        Delete the facebook object.

        :param object_id: ID for the facebook object(page,user..and so on)
        :param kwargs: Additional parameters.
        :return: Delete status.
        """
        resp = self._request(
            url=f"{self.version}/{object_id}",
            verb="DELETE",
            **kwargs,
        )
        data = self._parse_response(resp)
        return data

    def _get_oauth_session(
        self,
        redirect_uri: Optional[str] = None,
        scope: Optional[List[str]] = None,
        **kwargs,
    ) -> OAuth2Session:
        """
        :param redirect_uri: The URL that you want to redirect the person logging in back to.
        :param scope: A list of permission string to request from the person using your app.
        :param kwargs: Additional parameters for oauth.
        :return: OAuth Session
        """
        # check app credentials
        if not all([self.app_id, self.app_secret]):
            raise LibraryError({"message": "OAuth need your app credentials"})

        if redirect_uri is None:
            redirect_uri = self.DEFAULT_REDIRECT_URI
        if scope is None:
            scope = self.DEFAULT_SCOPE

        session = OAuth2Session(
            client_id=self.app_id,
            scope=scope,
            redirect_uri=redirect_uri,
            state=self.STATE,
            **kwargs,
        )
        session = facebook_compliance_fix(session)
        return session

    def get_authorization_url(
        self,
        redirect_uri: Optional[str] = None,
        scope: Optional[List[str]] = None,
        **kwargs,
    ) -> Tuple[str, str]:
        """
        Build authorization url to do oauth.
        Refer: https://developers.facebook.com/docs/facebook-login/manually-build-a-login-flow

        :param redirect_uri: The URL that you want to redirect the person logging in back to.
            Note: Your redirect uri need be set to `Valid OAuth redirect URIs` items in App Dashboard.
        :param scope: A list of permission string to request from the person using your app.
        :param kwargs: Additional parameters for oauth.
        :return: URL to do oauth and state
        """
        session = self._get_oauth_session(
            redirect_uri=redirect_uri, scope=scope, **kwargs
        )
        authorization_url, state = session.authorization_url(url=self.authorization_url)
        return authorization_url, state

    def exchange_user_access_token(
        self, response: str, redirect_uri: Optional[str] = None
    ) -> dict:
        """
        :param response: The redirect response url for authorize redirect
        :param redirect_uri: Url for your redirect.
        :return:
        """
        session = self._get_oauth_session(redirect_uri=redirect_uri)

        session.fetch_token(
            self.access_token_url,
            client_secret=self.app_secret,
            authorization_response=response,
        )
        self.access_token = session.access_token

        return session.token

    def exchange_page_access_token(
        self, page_id: str, access_token: Optional[str] = None
    ) -> str:
        """
        Get page access token by page administrator's user access token.

        Refer:
            1. https://developers.facebook.com/docs/pages/access-tokens
            2. https://developers.facebook.com/docs/facebook-login/access-tokens

        :param page_id: ID for page.
        :param access_token: Access token for user.
        :return: Page access token
        """
        if access_token is None:
            access_token = self.access_token

        resp = self._request(
            url=f"{self.version}/{page_id}",
            args={"fields": "access_token", "access_token": access_token},
            auth_need=False,
        )

        data = self._parse_response(resp)
        if "access_token" not in data:
            raise LibraryError(
                {
                    "message": "Can not get page access token. Reason maybe: \n"
                    "1. Your user access token has `page_show_list` or `manage_pages` permission.\n"
                    "2. You have the target page's manage permission."
                }
            )
        return data["access_token"]

    def exchange_long_lived_user_access_token(self, access_token=None) -> dict:
        """
        Generate long-lived token by short-lived token, Long-lived token generally lasts about 60 days.

        :param access_token: Short-lived user access token
        :return: Long-lived user access token info.
        """
        if access_token is None:
            access_token = self.access_token
        args = {
            "grant_type": "fb_exchange_token",
            "client_id": self.app_id,
            "client_secret": self.app_secret,
            "fb_exchange_token": access_token,
        }

        resp = self._request(
            url=self.access_token_url,
            args=args,
            auth_need=False,
        )
        data = self._parse_response(resp)
        return data

    def exchange_long_lived_page_access_token(
        self, user_id: str, access_token: Optional[str] = None
    ) -> dict:
        """
        Generate long-lived page access token by long-lived user access token.

        :param user_id: ID for the token user.
        :param access_token: Long-lived user token.
        :return: Data for Long-lived page token
        """

        data = self.get_connection(
            object_id=user_id,
            connection="accounts",
            access_token=access_token,
        )
        return data

    def get_app_token(
        self, app_id: Optional[str] = None, app_secret: Optional[str] = None
    ) -> dict:
        """
        Generate
        :param app_id: ID for app
        :param app_secret:  Secret for app
        :return: app access token
        """
        if app_id is None:
            app_id = self.app_id
        if app_secret is None:
            app_secret = self.app_secret

        resp = self._request(
            url=self.access_token_url,
            args={
                "grant_type": "client_credentials",
                "client_id": app_id,
                "client_secret": app_secret,
            },
            auth_need=False,
        )
        data = self._parse_response(resp)
        return data

    def debug_token(self, input_token: str, access_token: Optional[str] = None) -> dict:
        """
        Get input token info by issued app administrator's access token
        :param input_token: Access token to debug.
        :param access_token:
        :return: Issued app administrator's access token
        """
        if access_token is None:
            access_token = self.access_token

        resp = self._request(
            url=f"{self.version}/debug_token",
            args={"input_token": input_token, "access_token": access_token},
            auth_need=False,
        )
        data = self._parse_response(resp)
        return data


class BasicDisplayAPI(GraphAPI):
    GRAPH_URL = "https://graph.instagram.com/"
    AUTHORIZATION_URL = "https://api.instagram.com/oauth/authorize"
    EXCHANGE_ACCESS_TOKEN_URL = "https://api.instagram.com/oauth/access_token"

    DEFAULT_SCOPE = ["user_profile", "user_media"]

    def __init__(
        self,
        app_id: Optional[str] = None,
        app_secret: Optional[str] = None,
        access_token: Optional[str] = None,
        oauth_flow: bool = False,
        version: Optional[str] = None,
        sleep_on_rate_limit: bool = True,
        sleep_seconds_mapping: Optional[Dict[int, int]] = None,
        timeout: Optional[int] = None,
        proxies: Optional[dict] = None,
        base_url: Optional[str] = None,
        authorization_url: Optional[str] = None,
        access_token_url: Optional[str] = None,
    ):
        super().__init__(
            app_id=app_id,
            app_secret=app_secret,
            access_token=access_token,
            oauth_flow=oauth_flow,
            version=version,
            sleep_on_rate_limit=sleep_on_rate_limit,
            sleep_seconds_mapping=sleep_seconds_mapping,
            timeout=timeout,
            proxies=proxies,
            base_url=base_url,
            authorization_url=authorization_url,
            access_token_url=access_token_url,
        )

    @staticmethod
    def _generate_secret_proof(
        access_token: str, secret: Optional[str] = None
    ) -> Optional[str]:
        """
        :param access_token: Access token
        :param secret: App secret
        :return:
        """
        return None

    def exchange_user_access_token(
        self, response: str, redirect_uri: Optional[str] = None
    ) -> dict:
        """
        :param response: The redirect response url for authorize redirect
        :param redirect_uri: Url for your redirect.
        :return:
        """
        session = self._get_oauth_session(redirect_uri=redirect_uri)

        session.fetch_token(
            self.access_token_url,
            client_secret=self.app_secret,
            authorization_response=response,
            include_client_id=True,
        )
        self.access_token = session.access_token

        return session.token

    def exchange_long_lived_user_access_token(self, access_token=None) -> dict:
        """
        Exchange short-lived Instagram User Access Tokens for long-lived Instagram User Access Tokens.
        :param access_token: short-lived user token.
        :return: Long-lived user access token info.
        """
        if access_token is None:
            access_token = self.access_token

        args = {
            "grant_type": "ig_exchange_token",
            "client_secret": self.app_secret,
            "access_token": access_token,
        }
        resp = self._request(
            url=f"access_token",
            args=args,
            auth_need=False,
        )
        data = self._parse_response(resp)
        return data

    def refresh_access_token(self, access_token: str):
        """
        :param access_token: The valid (unexpired) long-lived Instagram User Access Token that you want to refresh.
        :return: New access token.
        """
        args = {"grant_type": "ig_refresh_token", "access_token": access_token}
        resp = self._request(
            url="refresh_access_token",
            args=args,
        )
        data = self._parse_response(resp)
        return data

    def exchange_page_access_token(
        self, page_id: str, access_token: Optional[str] = None
    ) -> str:
        raise LibraryError({"message": "Method not support"})

    def exchange_long_lived_page_access_token(
        self, user_id: str, access_token: Optional[str] = None
    ) -> dict:
        raise LibraryError({"message": "Method not support"})

    def get_app_token(
        self, app_id: Optional[str] = None, app_secret: Optional[str] = None
    ) -> dict:
        raise LibraryError({"message": "Method not support"})

    def debug_token(self, input_token: str, access_token: Optional[str] = None) -> dict:
        raise LibraryError({"message": "Method not support"})



"""
    Tests for comment.
"""

import responses


def test_get_info(helpers, fb_api):
    cm_id = "10158371815748553_10158371930463553"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{cm_id}",
            json=helpers.load_json(
                "testdata/facebook/apidata/comments/comment_info.json"
            ),
        )

        comment = fb_api.comment.get_info(comment_id=cm_id)
        assert comment.id == cm_id
        assert comment.created_time == "2021-07-21T21:05:14+0000"

        comment_json = fb_api.comment.get_info(
            comment_id=cm_id,
            fields="id,comment_count,created_time,like_count,message,permalink_url",
            return_json=True,
        )
        assert comment_json["id"] == cm_id


def test_get_batch(helpers, fb_api):
    cm_ids = [
        "10158371815748553_10158371930463553",
        "10158371815748553_10158372268858553",
    ]

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}",
            json=helpers.load_json(
                "testdata/facebook/apidata/comments/comments_info.json"
            ),
        )

        comments = fb_api.comment.get_batch(ids=cm_ids)
        assert comments[cm_ids[0]].id == cm_ids[0]

        comments_json = fb_api.comment.get_batch(
            ids=cm_ids,
            fields="id,comment_count,created_time,like_count,message,permalink_url",
            return_json=True,
        )
        assert comments_json[cm_ids[1]]["id"] == cm_ids[1]



"""
    Models for post.

    Refer: https://developers.facebook.com/docs/graph-api/reference/post/
"""

from dataclasses import dataclass
from typing import List, Optional

from dataclasses_json import config

from pyfacebook.models.base import BaseModel, field
from pyfacebook.models.extensions import Paging, Privacy, Reactions
from pyfacebook.models.attachment import Attachments
from pyfacebook.models.place import Place


@dataclass
class PostAction(BaseModel):
    """
    A class representing the post action
    """

    name: Optional[str] = field(repr=True)
    link: Optional[str] = field()


@dataclass
class PostTag(BaseModel):
    """
    A class representing the post tag.
    """

    id: Optional[str] = field(repr=True, compare=True)
    name: Optional[str] = field(repr=True)
    type: Optional[str] = field()
    offset: Optional[int] = field()
    length: Optional[int] = field()


@dataclass
class PostShare(BaseModel):
    """
    A class representing the post share
    """

    count: Optional[int] = field(repr=True)


@dataclass
class Post(BaseModel):
    """
    A class representing the post.
    """

    id: Optional[str] = field(repr=True, compare=True)
    actions: Optional[List[PostAction]] = field()
    admin_creator: Optional[dict] = field()  # TODO
    allowed_advertising_objectives: Optional[List[str]] = field()
    application: Optional[
        dict
    ] = (
        field()
    )  # TODO https://developers.facebook.com/docs/graph-api/reference/application/
    backdated_time: Optional[str] = field()
    call_to_action: Optional[dict] = field()
    can_reply_privately: Optional[bool] = field()
    child_attachments: Optional[List[dict]] = field()  # TODO
    comments_mirroring_domain: Optional[str] = field()
    coordinates: Optional[List[dict]] = field()  # TODO
    created_time: Optional[str] = field()
    event: Optional[
        dict
    ] = field()  # TODO https://developers.facebook.com/docs/graph-api/reference/event/
    expanded_height: Optional[int] = field()
    expanded_width: Optional[int] = field()
    feed_targeting: Optional[dict] = field()  # TODO
    _from: Optional[dict] = field(metadata=config(field_name="from"))
    full_picture: Optional[str] = field()
    height: Optional[int] = field()
    icon: Optional[str] = field()
    instagram_eligibility: Optional[str] = field()
    is_app_share: Optional[bool] = field()
    is_eligible_for_promotion: Optional[bool] = field()
    is_expired: Optional[bool] = field()
    is_hidden: Optional[bool] = field()
    is_inline_created: Optional[bool] = field()
    is_instagram_eligible: Optional[bool] = field()
    is_popular: Optional[bool] = field()
    is_published: Optional[bool] = field()
    is_spherical: Optional[bool] = field()
    message: Optional[str] = field(repr=True)
    message_tags: Optional[List[PostTag]] = field()
    multi_share_end_card: Optional[bool] = field()
    multi_share_optimized: Optional[bool] = field()
    parent_id: Optional[str] = field()
    permalink_url: Optional[str] = field()
    picture: Optional[str] = field()
    place: Optional[Place] = field()
    privacy: Optional[Privacy] = field()
    promotable_id: Optional[str] = field()
    properties: Optional[List[dict]] = field()  # TODO
    scheduled_publish_time: Optional[float] = field()
    shares: Optional[PostShare] = field()
    status_type: Optional[str] = field()
    story: Optional[str] = field()
    story_tags: Optional[List[PostTag]] = field()
    subscribed: Optional[bool] = field()
    target: Optional[dict] = field()
    targeting: Optional[dict] = field()  # TODO
    tagged_time: Optional[str] = field()  # Only for tagged posts.
    timeline_visibility: Optional[str] = field()
    updated_time: Optional[str] = field()
    via: Optional[dict] = field()
    video_buying_eligibility: Optional[List[str]] = field()
    width: Optional[int] = field()

    # common connections
    attachments: Optional[Attachments] = field()
    comments: Optional[dict] = field()  # TODO
    reactions: Optional[Reactions] = field()
    like: Optional[Reactions] = field()
    love: Optional[Reactions] = field()
    wow: Optional[Reactions] = field()
    haha: Optional[Reactions] = field()
    sad: Optional[Reactions] = field()
    angry: Optional[Reactions] = field()

    def __repr__(self):
        # Message shorten
        message = f"{self.message[:30]}..." if self.message else ""
        return f"Post(id={self.id},message={message})"


@dataclass
class FeedResponse(BaseModel):
    """
    A class representing the post response.

    Refer: https://developers.facebook.com/docs/graph-api/reference/page/feed/#read
    """

    data: List[Post] = field(repr=True, compare=True)
    paging: Optional[Paging] = field()



"""
    Models for place relative.

    Refer: https://developers.facebook.com/docs/graph-api/reference/place/
"""

from dataclasses import dataclass
from typing import Optional

from pyfacebook.models.base import BaseModel, field


@dataclass
class Location(BaseModel):
    city: Optional[str] = field()
    city_id: Optional[str] = field()
    country: Optional[str] = field(repr=True)
    country_code: Optional[str] = field()
    latitude: Optional[int] = field()
    located_in: Optional[str] = field()
    longitude: Optional[float] = field()
    name: Optional[str] = field()
    region: Optional[str] = field()
    region_id: Optional[int] = field()
    state: Optional[str] = field()
    street: Optional[str] = field()
    zip: Optional[str] = field()


@dataclass
class Place(BaseModel):
    """
    A class representing the Place.
    """

    id: Optional[str] = field(repr=True, compare=True)
    location: Optional[Location] = field()
    name: Optional[str] = field(repr=True)
    overall_rating: Optional[float] = field()



"""
    Tests for videos.
"""

import responses


def test_get_info(helpers, fb_api):
    video_id = "1192957457884299"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{video_id}",
            json=helpers.load_json(
                "testdata/facebook/apidata/videos/video_info_fields.json"
            ),
        )
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{video_id}",
            json=helpers.load_json(
                "testdata/facebook/apidata/videos/video_info_default.json"
            ),
        )

        video = fb_api.video.get_info(video_id=video_id)
        assert video.id == video_id
        assert video.updated_time == "2021-07-20T01:55:06+0000"

        video_json = fb_api.video.get_info(
            video_id=video_id,
            fields="id,description,updated_time",
            return_json=True,
        )
        assert video_json["id"] == video_id


def test_get_batch(helpers, fb_api):
    video_ids = ["1192957457884299", "334712884667245"]

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}",
            json=helpers.load_json(
                "testdata/facebook/apidata/videos/videos_info_fields.json"
            ),
        )
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}",
            json=helpers.load_json(
                "testdata/facebook/apidata/videos/videos_info_default.json"
            ),
        )

        videos = fb_api.video.get_batch(ids=video_ids)
        assert videos[video_ids[0]].id == video_ids[0]

        videos_json = fb_api.video.get_batch(
            ids=video_ids,
            fields="id,name,created_time",
            return_json=True,
        )
        assert videos_json[video_ids[1]]["id"] == video_ids[1]



"""
    Tests for message.
"""

import responses


def test_get_info(helpers, fb_api):
    msg_id = "m_ToF35NI1OImBjyUIgplSaBMylUFmkYY4bHogy9C1otLISU6SGhccB5NK-THX_W4EdVQiKVv5SgCW9m-_C78mRA"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{msg_id}",
            json=helpers.load_json(
                "testdata/facebook/apidata/messages/message_info.json"
            ),
        )

        message = fb_api.message.get_info(message_id=msg_id)
        assert message.id == msg_id
        assert message.created_time == "2021-07-23T09:50:48+0000"

        message_json = fb_api.message.get_info(
            message_id=msg_id,
            fields="id,created_time,from,to,tags,attachments",
            return_json=True,
        )
        assert message_json["id"] == msg_id


def test_get_batch(helpers, fb_api):
    msg_ids = [
        "m_ToF35NI1OImBjyUIgplSaBMylUFmkYY4bHogy9C1otLISU6SGhccB5NK-THX_W4EdVQiKVv5SgCW9m-_C78mRA",
        "m_rU2w1v7XMh8JR7jDSmW3pBMylUFmkYY4bHogy9C1otJUI2R3JKjhRblmg6DHQeVLHU6GP0qwqSqdSg0gnkE8tw",
    ]

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}",
            json=helpers.load_json(
                "testdata/facebook/apidata/messages/messages_info.json"
            ),
        )

        messages = fb_api.message.get_batch(ids=msg_ids)
        assert messages[msg_ids[0]].id == msg_ids[0]

        messages_json = fb_api.message.get_batch(
            ids=msg_ids,
            fields="id,created_time,from,to,tags,attachments",
            return_json=True,
        )
        assert messages_json[msg_ids[1]]["id"] == msg_ids[1]



"""
    Apis for page.
"""
from typing import Dict, Optional, Union

import pyfacebook.utils.constant as const
from pyfacebook.api.base_resource import BaseResource
from pyfacebook.api.facebook.common_edges import (
    FeedEdge,
    AlbumsEdge,
    PhotosEdge,
    VideosEdge,
    LiveVideosEdge,
    CommentsEdge,
)
from pyfacebook.exceptions import LibraryError
from pyfacebook.models.page import Page, SearchPagesResponse
from pyfacebook.models.post import FeedResponse
from pyfacebook.utils.params_utils import enf_comma_separated


class FacebookPage(
    BaseResource,
    FeedEdge,
    AlbumsEdge,
    PhotosEdge,
    VideosEdge,
    LiveVideosEdge,
    CommentsEdge,
):
    def get_info(
        self,
        page_id: Optional[str] = None,
        username: Optional[str] = None,
        fields: Optional[Union[str, list, tuple]] = None,
        return_json: bool = False,
    ) -> Union[Page, dict]:
        """
        Get information about a Facebook Page.

        :param page_id: ID for the page.
        :param username: Username for the page.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a dataclass for page.
            Or return json data. Default is false.
        :return: Page information.
        """
        if page_id:
            target = page_id
        elif username:
            target = username
        else:
            raise LibraryError(
                {"message": "Specify at least one of page_id or username"}
            )

        if fields is None:
            fields = const.PAGE_PUBLIC_FIELDS

        data = self.client.get_object(
            object_id=target,
            fields=enf_comma_separated(field="fields", value=fields),
        )
        if return_json:
            return data
        else:
            return Page.new_from_json_dict(data=data)

    def get_batch(
        self,
        ids: Optional[Union[str, list, tuple]],
        fields: Optional[Union[str, list, tuple]] = None,
        return_json: bool = False,
    ) -> Union[Dict[str, Page], dict]:
        """
        Get batch pages information by ids.

        :param ids: IDs for the pages.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a dataclass for page.
            Or return json data. Default is false.
        :return: Pages information.
        """

        ids = enf_comma_separated(field="ids", value=ids)

        if fields is None:
            fields = const.PAGE_PUBLIC_FIELDS

        data = self.client.get_objects(
            ids=ids, fields=enf_comma_separated(field="fields", value=fields)
        )
        if return_json:
            return data
        else:
            return {
                page_id: Page.new_from_json_dict(item) for page_id, item in data.items()
            }

    def get_posts(
        self,
        object_id: str,
        fields: Optional[Union[str, list, dict]] = None,
        since: Optional[str] = None,
        until: Optional[str] = None,
        count: Optional[int] = 10,
        limit: Optional[int] = 10,
        return_json: bool = False,
    ) -> Union[FeedResponse, dict]:
        """
        Get the page's own posts.

        :param object_id: ID for page to get posts.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param since: A Unix timestamp or strtotime data value that points to the start of data.
        :param until: A Unix timestamp or strtotime data value that points to the end of data.
        :param count: The total count for you to get data.
        :param limit: Each request retrieve objects count.
            It should no more than 100. Default is None will use api default limit.
        :param return_json: Set to false will return a dataclass for post.
            Or return json data. Default is false.
        :return: Posts response information
        """
        return self._get_feed(
            object_id=object_id,
            fields=fields,
            since=since,
            until=until,
            count=count,
            limit=limit,
            source="posts",
            return_json=return_json,
        )

    def get_published_posts(
        self,
        object_id: str,
        fields: Optional[Union[str, list, dict]] = None,
        since: Optional[str] = None,
        until: Optional[str] = None,
        count: Optional[int] = 10,
        limit: Optional[int] = 10,
        return_json: bool = False,
    ) -> Union[FeedResponse, dict]:
        """
        Get all published posts by this page.

        :param object_id: ID for page to get published posts.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param since: A Unix timestamp or strtotime data value that points to the start of data.
        :param until: A Unix timestamp or strtotime data value that points to the end of data.
        :param count: The total count for you to get data.
        :param limit: Each request retrieve objects count.
            It should no more than 100. Default is None will use api default limit.
        :param return_json: Set to false will return a dataclass for post.
            Or return json data. Default is false.
        :return: Posts response information
        """
        return self._get_feed(
            object_id=object_id,
            fields=fields,
            since=since,
            until=until,
            count=count,
            limit=limit,
            source="published_posts",
            return_json=return_json,
        )

    def get_tagged_posts(
        self,
        object_id: str,
        fields: Optional[Union[str, list, dict]] = None,
        since: Optional[str] = None,
        until: Optional[str] = None,
        count: Optional[int] = 10,
        limit: Optional[int] = 10,
        return_json: bool = False,
    ) -> Union[FeedResponse, dict]:
        """
        Get posts which tagged the target page.

        :param object_id: ID for page to get tagged posts.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param since: A Unix timestamp or strtotime data value that points to the start of data.
        :param until: A Unix timestamp or strtotime data value that points to the end of data.
        :param count: The total count for you to get data.
        :param limit: Each request retrieve objects count.
            It should no more than 100. Default is None will use api default limit.
        :param return_json: Set to false will return a dataclass for post.
            Or return json data. Default is false.
        :return: Posts response information.
        """
        return self._get_feed(
            object_id=object_id,
            fields=fields,
            since=since,
            until=until,
            count=count,
            limit=limit,
            source="tagged",
            return_json=return_json,
        )

    def search(
        self,
        q: str,
        fields: Optional[Union[str, list, dict]] = None,
        count: Optional[int] = 25,
        limit: Optional[int] = 25,
        return_json: bool = False,
    ) -> Union[SearchPagesResponse, dict]:
        """
        Returns a list of Pages that meet the query's criteria.

        :param q: Value to a keyword or search term.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param count: The total count for you to get data.
        :param limit: Each request retrieve objects count.
            It should no more than 100. Default is None will use api default limit.
        :param return_json: Set to false will return a dataclass for page.
            Or return json data. Default is false.
        :return: Pages search response.
        """
        if fields is None:
            fields = const.SEARCH_PAGES_PUBLIC_FIELDS

        data = self.client.get_full_connections(
            object_id="pages",
            connection="search",
            count=count,
            limit=limit,
            q=q,
            fields=enf_comma_separated(field="fields", value=fields),
        )
        if return_json:
            return data
        else:
            return SearchPagesResponse.new_from_json_dict(data)



import pytest

from pyfacebook import IGBasicDisplayApi


@pytest.fixture
def api():
    return IGBasicDisplayApi(
        app_id="123456",
        app_secret="xxxxx",
        access_token="token",
    )



"""
    Model class for Message.

    Refer: https://developers.facebook.com/docs/graph-api/reference/message
"""

from dataclasses import dataclass
from typing import List, Optional

from dataclasses_json import config

from pyfacebook.models.base import BaseModel, field


@dataclass
class Tag(BaseModel):
    """
    A class representing the tag in message.
    """

    name: Optional[str] = field(repr=True)


@dataclass
class MessageTags(BaseModel):
    """
    A class representing the tags in message.
    """

    data: Optional[List[Tag]] = field(repr=True)


@dataclass
class MsgUser(BaseModel):
    """
    A class representing the User in message.
    """

    id: Optional[str] = field(repr=True, compare=True)
    name: Optional[str] = field(repr=True)
    email: Optional[str] = field()


@dataclass
class MessageTo(BaseModel):
    """
    A class representing the to in message.
    """

    data: Optional[List[MsgUser]] = field(repr=True)


@dataclass
class MessageAttachmentImageData(BaseModel):
    """
    A class representing the image in attachment.
    """

    width: Optional[int] = field()
    height: Optional[int] = field()
    max_width: Optional[int] = field()
    max_height: Optional[int] = field()
    url: Optional[str] = field(repr=True)
    preview_url: Optional[str] = field()
    raw_gif_image: Optional[str] = field()
    raw_webp_image: Optional[str] = field()
    animated_gif_url: Optional[str] = field()
    animated_gif_preview_url: Optional[str] = field()
    animated_webp_url: Optional[str] = field()
    animated_webp_preview_url: Optional[str] = field()
    image_type: Optional[int] = field()
    render_as_sticker: Optional[bool] = field()


@dataclass
class MessageAttachmentVideoData(BaseModel):
    """
    A class representing the video in attachment.
    """

    width: Optional[int] = field()
    height: Optional[int] = field()
    length: Optional[int] = field()
    video_type: Optional[int] = field()
    url: Optional[str] = field(repr=True)
    preview_url: Optional[str] = field()
    rotation: Optional[int] = field()


@dataclass
class MessageAttachment(BaseModel):
    """
    A class representing the Attachment in message.

    Refer: https://developers.facebook.com/docs/graph-api/reference/v11.0/message/attachments
    """

    id: Optional[str] = field(repr=True, compare=True)
    mime_type: Optional[str] = field()
    name: Optional[str] = field()
    size: Optional[int] = field()
    file_url: Optional[str] = field()
    image_data: Optional[MessageAttachmentImageData] = field()
    video_data: Optional[MessageAttachmentVideoData] = field()


@dataclass
class MessageAttachments(BaseModel):
    """
    A class representing the attachments in message.
    """

    data: Optional[List[MessageAttachment]] = field(repr=True)


@dataclass
class Message(BaseModel):
    """
    A class representing the Message
    """

    id: Optional[str] = field(repr=True, compare=True)
    created_time: Optional[str] = field()
    _from: Optional[dict] = field(metadata=config(field_name="from"))
    message: Optional[str] = field()
    tags: Optional[MessageTags] = field()
    to: Optional[MessageTo] = field()
    sticker: Optional[str] = field()

    # common connections
    attachments: Optional[MessageAttachments] = field()



"""
    Tests for page api
"""

import pytest
import responses

from pyfacebook.exceptions import LibraryError


def test_get_info(helpers, fb_api):
    pid = "20531316728"
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{pid}",
            json=helpers.load_json(
                "testdata/facebook/apidata/pages/single_fields_page.json"
            ),
        )

        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/facebookapp",
            json=helpers.load_json(
                "testdata/facebook/apidata/pages/single_default_page.json"
            ),
        )

        page = fb_api.page.get_info(page_id=pid)
        assert page.id == pid

        page_json = fb_api.page.get_info(
            username="facebookapp",
            fields="id,about,can_checkin,category,category_list,checkins,contact_address,cover,current_location,description,description_html,display_subtext,emails,engagement,fan_count,founded,general_info,global_brand_page_name,global_brand_root_id,link,name,phone,picture,rating_count,single_line_address,start_info,talking_about_count,username,verification_status,website,were_here_count,whatsapp_number",
            return_json=True,
        )
        assert page_json["id"] == pid

    with pytest.raises(LibraryError):
        fb_api.page.get_info()


def test_get_batch(helpers, fb_api):
    page_ids = ["20531316728", "19292868552"]

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}",
            json=helpers.load_json(
                "testdata/facebook/apidata/pages/multi_default_fields.json"
            ),
        )
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}",
            json=helpers.load_json("testdata/facebook/apidata/pages/multi_pages.json"),
        )

        data = fb_api.page.get_batch(ids=page_ids)
        assert page_ids[0] in data.keys()

        data_json = fb_api.page.get_batch(
            ids=page_ids, fields="id,name,username,fan_count", return_json=True
        )
        assert data_json[page_ids[0]]["id"] == page_ids[0]


def test_get_feed(helpers, fb_api):
    pid = "19292868552"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{pid}/feed",
            json=helpers.load_json(
                "testdata/facebook/apidata/posts/feeds_default_fields_p1.json"
            ),
        )
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{pid}/feed",
            json=helpers.load_json(
                "testdata/facebook/apidata/posts/feeds_default_fields_p2.json"
            ),
        )

        feed = fb_api.page.get_feed(object_id=pid, count=None, limit=5)
        assert len(feed.data) == 10
        assert feed.data[0].id == "19292868552_10158349356748553"


def test_get_posts(helpers, fb_api):
    pid = "19292868552"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{pid}/posts",
            json=helpers.load_json(
                "testdata/facebook/apidata/posts/feeds_default_fields_p1.json"
            ),
        )

        feed_json = fb_api.page.get_posts(
            object_id=pid, count=4, limit=5, return_json=True
        )
        assert len(feed_json["data"]) == 4
        assert feed_json["data"][0]["id"] == "19292868552_10158349356748553"


def test_get_published_posts(helpers, fb_api):
    pid = "19292868552"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{pid}/published_posts",
            json=helpers.load_json(
                "testdata/facebook/apidata/posts/feeds_default_fields_p1.json"
            ),
        )

        feed = fb_api.page.get_published_posts(object_id=pid, count=4)
        assert len(feed.data) == 4
        assert feed.data[0].id == "19292868552_10158349356748553"


def test_get_tagged_posts(helpers, fb_api):
    pid = "19292868552"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{pid}/tagged",
            json=helpers.load_json(
                "testdata/facebook/apidata/posts/feeds_default_fields_p1.json"
            ),
        )

        feed = fb_api.page.get_tagged_posts(object_id=pid, count=4)
        assert len(feed.data) == 4
        assert feed.data[0].id == "19292868552_10158349356748553"


def test_get_albums(helpers, fb_api):
    pid = "2121008874780932"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{pid}/albums",
            json=helpers.load_json(
                "testdata/facebook/apidata/albums/albums_list_p1.json"
            ),
        )
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{pid}/albums",
            json=helpers.load_json(
                "testdata/facebook/apidata/albums/albums_list_p2.json"
            ),
        )

        albums = fb_api.page.get_albums(object_id=pid, count=None, limit=3)
        assert len(albums.data) == 6
        assert albums.data[0].id == "2312974342251050"

        albums_json = fb_api.page.get_albums(
            object_id=pid, count=3, limit=3, return_json=True
        )
        assert len(albums_json["data"]) == 3


def test_get_photos(helpers, fb_api):
    pid = "108824017345866"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{pid}/photos",
            json=helpers.load_json("testdata/facebook/apidata/photos/photos_p1.json"),
        )
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{pid}/photos",
            json=helpers.load_json("testdata/facebook/apidata/photos/photos_p2.json"),
        )

        photos = fb_api.page.get_photos(object_id=pid, count=None, limit=2)
        assert len(photos.data) == 4
        assert photos.data[0].id == "336596487901950"

        photos_json = fb_api.page.get_photos(
            object_id=pid, count=2, limit=2, return_json=True
        )
        assert len(photos_json["data"]) == 2


def test_get_live_videos(helpers, fb_api):
    pid = "20531316728"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{pid}/live_videos",
            json=helpers.load_json(
                "testdata/facebook/apidata/live_videos/live_videos_p1.json"
            ),
        )
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{pid}/live_videos",
            json=helpers.load_json(
                "testdata/facebook/apidata/live_videos/live_videos_p2.json"
            ),
        )

        live_videos = fb_api.page.get_live_videos(object_id=pid, count=None, limit=3)
        assert len(live_videos.data) == 6
        assert live_videos.data[0].id == "10160659411121729"

        live_videos_json = fb_api.page.get_live_videos(
            object_id=pid, count=3, limit=3, return_json=True
        )
        assert len(live_videos_json["data"]) == 3


def test_get_videos(helpers, fb_api):
    pid = "19292868552"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{pid}/videos",
            json=helpers.load_json("testdata/facebook/apidata/videos/videos_p1.json"),
        )
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{pid}/videos",
            json=helpers.load_json("testdata/facebook/apidata/videos/videos_p2.json"),
        )

        videos = fb_api.page.get_videos(object_id=pid, count=None, limit=3)
        assert len(videos.data) == 5
        assert videos.data[0].id == "1002065083862711"

        videos_json = fb_api.page.get_videos(
            object_id=pid, count=2, limit=3, return_json=True
        )
        assert len(videos_json["data"]) == 2


def test_search_pages(helpers, fb_api):
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/pages/search",
            json=helpers.load_json("testdata/facebook/apidata/pages/search_pages.json"),
        )

        pages = fb_api.page.search(q="facebook", count=5, limit=5)
        assert len(pages.data) == 5
        assert pages.data[0].id == "108824017345866"

        pages_json = fb_api.page.search(q="facebook", count=4, return_json=True)
        assert len(pages_json["data"]) == 4



"""
    Tests for business
"""

import responses


def test_get_info(helpers, fb_api):
    b_id = "123456789"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{b_id}",
            json=helpers.load_json(
                "testdata/facebook/apidata/business/business_info.json"
            ),
        )

        business = fb_api.business.get_info(business_id=b_id)
        assert business.id == b_id
        assert business.updated_time == "2021-08-31T08:27:35+0000"

        business_json = fb_api.business.get_info(
            business_id=b_id,
            fields="id,name,link,created_time,updated_time,verification_status,profile_picture_uri",
            return_json=True,
        )
        assert business_json["id"] == b_id


def test_get_batch(helpers, fb_api):
    b_ids = ["123456789", "987654321"]

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}",
            json=helpers.load_json(
                "testdata/facebook/apidata/business/businesses_info.json"
            ),
        )

        businesses = fb_api.business.get_batch(ids=b_ids)
        assert businesses[b_ids[0]].id == b_ids[0]

        business_json = fb_api.business.get_batch(
            ids=b_ids,
            fields="id,name,link,created_time,updated_time,verification_status,profile_picture_uri",
            return_json=True,
        )
        assert business_json[b_ids[1]]["id"] == b_ids[1]



"""
    Apis for basic user
"""
from typing import Optional, Union

import pyfacebook.utils.constant as const
from pyfacebook.api.base_resource import BaseResource
from pyfacebook.models.ig_basic_models import IgBasicUser, IgBasicMediaResponse
from pyfacebook.utils.params_utils import enf_comma_separated


class IGBasicUser(BaseResource):
    def get_info(
        self,
        user_id: Optional[str] = "me",
        fields: Optional[Union[str, list, tuple]] = None,
        return_json: bool = False,
    ) -> Union[IgBasicUser, dict]:
        """
        Represents an Instagram user profile.

        :param user_id: ID for the user, matched the access token.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a dataclass for Account.
            Or return json data. Default is false.
        :return: IG Basic User information.
        """
        if fields is None:
            fields = const.IG_BASIC_USER_FIELDS

        data = self.client.get_object(
            object_id=user_id,
            fields=enf_comma_separated(field="fields", value=fields),
        )
        if return_json:
            return data
        else:
            return IgBasicUser.new_from_json_dict(data=data)

    def get_media(
        self,
        user_id: Optional[str] = "me",
        fields: Optional[Union[str, list, tuple]] = None,
        since: Optional[str] = None,
        until: Optional[str] = None,
        count: Optional[int] = 10,
        limit: Optional[int] = 10,
        return_json=False,
    ) -> Union[IgBasicMediaResponse, dict]:
        """
        Represents a collection of Media on a User. Maximum of 10K of the most recently created media.

        :param user_id: ID for the user, matched the access token.
        :param fields:
        :param since: A Unix timestamp or strtotime data value that points to the start of data.
        :param until: A Unix timestamp or strtotime data value that points to the end of data.
        :param count: The total count for you to get data.
        :param limit: Each request retrieve objects count.
            It should no more than 100. Default is None will use api default limit.
        :param return_json: Set to false will return a dataclass for IgBasicMediaResponse.
            Or return json data. Default is false.
        :return: Media response information.
        """

        if fields is None:
            fields = const.IG_BASIC_MEDIA_FIELDS

        data = self.client.get_full_connections(
            object_id=user_id,
            connection="media",
            count=count,
            limit=limit,
            fields=enf_comma_separated(field="fields", value=fields),
            since=since,
            until=until,
        )
        if return_json:
            return data
        else:
            return IgBasicMediaResponse.new_from_json_dict(data)



"""
    Tests for groups
"""

import responses


def test_get_info(helpers, fb_api):
    gp_id = "251560641854558"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{gp_id}",
            json=helpers.load_json(
                "testdata/facebook/apidata/groups/group_default_fields.json"
            ),
        )
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{gp_id}",
            json=helpers.load_json(
                "testdata/facebook/apidata/groups/group_fields.json"
            ),
        )

        gp = fb_api.group.get_info(group_id=gp_id)
        assert gp.id == gp_id
        assert gp.member_count == 188310

        gp_json = fb_api.group.get_info(
            group_id=gp_id,
            fields="id,name,description,created_time,member_count",
            return_json=True,
        )
        assert gp_json["id"] == gp_id


def test_get_batch(helpers, fb_api):
    gp_ids = ["2260975870792283", "251560641854558"]

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}",
            json=helpers.load_json(
                "testdata/facebook/apidata/groups/groups_default_fields.json"
            ),
        )
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}",
            json=helpers.load_json(
                "testdata/facebook/apidata/groups/groups_fields.json"
            ),
        )

        gps = fb_api.group.get_batch(ids=gp_ids)
        assert gps[gp_ids[0]].id == gp_ids[0]

        gps_json = fb_api.group.get_batch(
            ids=gp_ids,
            fields="id,name,description,created_time,member_count",
            return_json=True,
        )
        assert gps_json[gp_ids[1]]["id"] == gp_ids[1]


def test_get_feed(helpers, fb_api):
    gp_id = "124"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{gp_id}/feed",
            json=helpers.load_json("testdata/facebook/apidata/groups/feed_p1.json"),
        )
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{gp_id}/feed",
            json=helpers.load_json("testdata/facebook/apidata/groups/feed_p2.json"),
        )

        feed = fb_api.group.get_feed(object_id=gp_id)
        assert len(feed.data) == 10



"""
    Models for group.

    Refer: https://developers.facebook.com/docs/graph-api/reference/group
"""

from dataclasses import dataclass
from typing import List, Optional

from pyfacebook.models.base import BaseModel, field
from pyfacebook.models.image import CoverPhoto, Picture


@dataclass
class Group(BaseModel):
    """
    A class representing the group.

    Refer: https://developers.facebook.com/docs/graph-api/reference/group
    """

    id: Optional[str] = field(repr=True, compare=True)
    cover: Optional[CoverPhoto] = field()
    created_time: Optional[str] = field()
    description: Optional[str] = field()
    email: Optional[str] = field()
    icon: Optional[str] = field()
    member_count: Optional[int] = field()
    member_request_count: Optional[int] = field()
    membership_state: Optional[str] = field()
    name: Optional[str] = field()
    parent: Optional[dict] = field()  # TODO
    permissions: Optional[List[str]] = field()
    privacy: Optional[str] = field()
    updated_time: Optional[str] = field()

    # common connection
    picture: Optional[Picture] = field()



"""
    Tests for hashtag.
"""

import responses


def test_get_info(helpers, api):
    hashtag_id = "17843826142012701"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{hashtag_id}",
            json=helpers.load_json(
                "testdata/instagram/apidata/hashtags/hashtag_info.json"
            ),
        )

        hashtag = api.hashtag.get_info(hashtag_id=hashtag_id)
        assert hashtag.id == hashtag_id

        hashtag_json = api.hashtag.get_info(
            hashtag_id=hashtag_id,
            fields="id,name",
            return_json=True,
        )
        assert hashtag_json["id"] == hashtag_id


def test_get_batch(helpers, api):
    hashtag_ids = ["17843826142012701", "17841593698074073"]

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}",
            json=helpers.load_json(
                "testdata/instagram/apidata/hashtags/hashtags_info.json"
            ),
        )

        hashtags = api.hashtag.get_batch(ids=hashtag_ids)
        assert hashtags[hashtag_ids[0]].id == hashtag_ids[0]

        hashtags_json = api.hashtag.get_batch(
            ids=hashtag_ids,
            fields="id,name",
            return_json=True,
        )
        assert hashtags_json[hashtag_ids[0]]["id"] == hashtag_ids[0]


def test_get_top_media(helpers, api):
    hashtag_id = "17841562426109234"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{hashtag_id}/top_media",
            json=helpers.load_json(
                "testdata/instagram/apidata/hashtags/hashtag_top_medias_p1.json"
            ),
        )
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{hashtag_id}/top_media",
            json=helpers.load_json(
                "testdata/instagram/apidata/hashtags/hashtag_top_medias_p2.json"
            ),
        )

        top_media = api.hashtag.get_top_media(
            hashtag_id=hashtag_id,
            count=None,
            limit=25,
        )
        assert len(top_media.data) == 50

        top_media_json = api.hashtag.get_top_media(
            hashtag_id=hashtag_id,
            count=10,
            return_json=True,
        )
        assert len(top_media_json["data"]) == 10


def test_get_recent_media(helpers, api):
    hashtag_id = "17841562426109234"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{hashtag_id}/recent_media",
            json=helpers.load_json(
                "testdata/instagram/apidata/hashtags/hashtag_recent_medias_p1.json"
            ),
        )
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{hashtag_id}/recent_media",
            json=helpers.load_json(
                "testdata/instagram/apidata/hashtags/hashtag_recent_medias_p2.json"
            ),
        )

        top_media = api.hashtag.get_recent_media(
            hashtag_id=hashtag_id,
            count=None,
            limit=5,
        )
        assert len(top_media.data) == 10

        top_media_json = api.hashtag.get_recent_media(
            hashtag_id=hashtag_id,
            count=5,
            return_json=True,
        )
        assert len(top_media_json["data"]) == 5



"""
    Tests for event.
"""

import responses


def test_get_info(helpers, fb_api):
    evt_id = "5971414199599456"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{evt_id}",
            json=helpers.load_json("testdata/facebook/apidata/events/event_info.json"),
        )

        event = fb_api.event.get_info(event_id=evt_id)
        assert event.id == evt_id
        assert event.start_time == "2021-07-23T21:00:00+0800"

        event_json = fb_api.event.get_info(
            event_id=evt_id,
            fields="id,name,description,start_time,end_time,place,event_times",
            return_json=True,
        )
        assert event_json["id"] == evt_id


def test_get_batch(helpers, fb_api):
    et_ids = ["5971414199599456", "512411373205445"]

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}",
            json=helpers.load_json("testdata/facebook/apidata/events/events_info.json"),
        )

        events = fb_api.event.get_batch(ids=et_ids)
        assert events[et_ids[0]].id == et_ids[0]

        events_json = fb_api.event.get_batch(
            ids=et_ids,
            fields="id,name,description,start_time,end_time,place,event_times",
            return_json=True,
        )
        assert events_json[et_ids[1]]["id"] == et_ids[1]



"""
    Exceptions for library
"""


class PyFacebookException(Exception):
    """Base class for exceptions in this module."""

    pass


class LibraryError(PyFacebookException):
    """
    A class for library error template.
    """

    def __init__(self, kwargs: dict):
        self.code = -1  # error code in library inside
        self.message = "exception in library"
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __repr__(self):
        msg = ",".join(
            [f"{k}={v}" for k, v in self.__dict__.items() if not k.startswith("_")]
        )
        return f"{self.__class__.__name__}({msg})"

    def __str__(self):
        return self.__repr__()


class FacebookError(LibraryError):
    """
    A class representing facebook error response
    Refer: https://developers.facebook.com/docs/graph-api/using-graph-api/error-handling

    Common fields are: message,code,error_subcode,error_user_msg,error_user_title,fbtrace_id
    """

    def __init__(self, kwargs: dict):
        self._data = kwargs
        error = kwargs["error"]
        super().__init__(kwargs=error)


class PyFacebookDeprecationWaring(DeprecationWarning):
    pass



"""
    Base client for API.
"""
import inspect

from pyfacebook import GraphAPI, BasicDisplayAPI
from pyfacebook.api.base_resource import BaseResource


def _is_resource_cls(obj):
    return isinstance(obj, BaseResource)


class BaseApi(GraphAPI):
    def __new__(cls, *args, **kwargs):
        self = super().__new__(cls)
        resources = inspect.getmembers(cls, _is_resource_cls)
        for name, resource in resources:
            resource_cls = type(resource)
            resource = resource_cls(self)
            setattr(self, name, resource)
        return self


class BaseBasicDisplayApi(BasicDisplayAPI, BaseApi):
    ...



"""
    Tests for publishes.
"""

import responses


def test_get_info(helpers, api):
    container_id = "17966279578433554"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{container_id}",
            json=helpers.load_json(
                "testdata/instagram/apidata/publish/container_info.json"
            ),
        )

        container = api.container.get_info(container_id=container_id)
        assert container.id == container_id

        container_json = api.container.get_info(
            container_id=container_id,
            fields="id,status_code",
            return_json=True,
        )
        assert container_json["id"] == container_id


def test_get_batch(helpers, api):
    container_ids = ["17883019676403955", "17966279578433554"]

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}",
            json=helpers.load_json(
                "testdata/instagram/apidata/publish/containers_info.json"
            ),
        )

        containers = api.container.get_batch(ids=container_ids)
        assert containers[container_ids[0]].id == container_ids[0]

        containers_json = api.container.get_batch(
            ids=container_ids,
            fields="id,status_code",
            return_json=True,
        )
        assert containers_json[container_ids[0]]["id"] == container_ids[0]



"""
    Apis for User.
"""
from typing import Dict, Optional, Union

import pyfacebook.utils.constant as const
from pyfacebook.api.base_resource import BaseResource
from pyfacebook.api.facebook.common_edges import (
    FeedEdge,
    AlbumsEdge,
    PhotosEdge,
    VideosEdge,
    LiveVideosEdge,
)
from pyfacebook.models.user import User
from pyfacebook.models.page import PagesResponse
from pyfacebook.models.post import FeedResponse
from pyfacebook.models.business import BusinessResponse
from pyfacebook.utils.params_utils import enf_comma_separated


class FacebookUser(
    BaseResource, FeedEdge, AlbumsEdge, PhotosEdge, VideosEdge, LiveVideosEdge
):
    def get_info(
        self,
        user_id: str,
        fields: Optional[Union[str, list, tuple]] = None,
        return_json: bool = False,
    ) -> Union[User, dict]:
        """
        Get information about a Facebook User.

        :param user_id: ID for user.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a dataclass for user.
        Or return json data. Default is false.
        :return: User information.
        """
        if fields is None:
            fields = const.USER_PUBLIC_FIELDS

        data = self.client.get_object(
            object_id=user_id,
            fields=enf_comma_separated(field="fields", value=fields),
        )
        if return_json:
            return data
        else:
            return User.new_from_json_dict(data=data)

    def get_batch(
        self,
        ids: Optional[Union[str, list, tuple]],
        fields: Optional[Union[str, list, tuple]] = None,
        return_json: bool = False,
    ) -> Union[Dict[str, User], dict]:
        """
        Get batch users information by ids

        :param ids: IDs for the users.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a dataclass for user.
            Or return json data. Default is false.
        :return: Users information.
        """
        ids = enf_comma_separated(field="ids", value=ids)

        if fields is None:
            fields = const.USER_PUBLIC_FIELDS

        data = self.client.get_objects(
            ids=ids, fields=enf_comma_separated(field="fields", value=fields)
        )
        if return_json:
            return data
        else:
            return {
                user_id: User.new_from_json_dict(item) for user_id, item in data.items()
            }

    def get_accounts(
        self,
        user_id: str,
        fields: Optional[Union[str, list, tuple]] = None,
        is_place: Optional[bool] = None,
        is_promotable: Optional[bool] = None,
        summary: Optional[bool] = None,
        count: Optional[int] = 10,
        limit: Optional[int] = 10,
        return_json: bool = False,
    ) -> Union[PagesResponse, dict]:
        """
        Get the Facebook Pages that a person owns or is able to perform tasks on.

        :param user_id: ID for the user.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param is_place: If specified,filter pages based on whetherthey are places or not.
        :param is_promotable: If specified, filter pages based on whether they can be promoted or not.
        :param summary: Whether aggregated information about the edge, such as counts.
        :param count: The total count for you to get data.
        :param limit: Each request retrieve objects count.
            It should no more than 100. Default is None will use api default limit.
        :param return_json: Set to false will return a dataclass for Page.
            Or return json data. Default is false.
        :return: Pages information and paging
        """
        if fields is None:
            fields = const.PAGE_PUBLIC_FIELDS

        data = self.client.get_full_connections(
            object_id=user_id,
            connection="accounts",
            count=count,
            limit=limit,
            fields=enf_comma_separated(field="fields", value=fields),
            is_place=is_place,
            is_promotable=is_promotable,
            summary=summary,
        )
        if return_json:
            return data
        else:
            return PagesResponse.new_from_json_dict(data)

    def get_posts(
        self,
        object_id: str,
        fields: Optional[Union[str, list, dict]] = None,
        since: Optional[str] = None,
        until: Optional[str] = None,
        count: Optional[int] = 10,
        limit: Optional[int] = 10,
        return_json: bool = False,
    ) -> Union[FeedResponse, dict]:
        """
        Get the posts published by the person themselves.

        :param object_id: ID for the user.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param since: A Unix timestamp or strtotime data value that points to the start of data.
        :param until: A Unix timestamp or strtotime data value that points to the end of data.
        :param count: The total count for you to get data.
        :param limit: Each request retrieve objects count.
            It should no more than 100. Default is None will use api default limit.
        :param return_json: Set to false will return a dataclass for post.
            Or return json data. Default is false.
        :return: Posts information and paging
        """
        return self._get_feed(
            object_id=object_id,
            fields=fields,
            since=since,
            until=until,
            count=count,
            limit=limit,
            source="posts",
            return_json=return_json,
        )

    def get_businesses(
        self,
        user_id: str,
        fields: Optional[Union[str, list, dict]] = None,
        count: Optional[int] = 10,
        limit: Optional[int] = 10,
        return_json: bool = False,
    ) -> Union[BusinessResponse, dict]:
        """
        Get the businesses for the user.

        :param user_id: ID for the user.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param count: The total count for you to get data.
        :param limit: Each request retrieve objects count.
            It should no more than 100. Default is None will use api default limit.
        :param return_json: Set to false will return a list of dataclass for Businesses.
            Or return json data. Default is false.
        :return: Response for the user businesses.
        """

        if fields is None:
            fields = const.BUSINESS_PUBLIC_FIELDS

        data = self.client.get_full_connections(
            object_id=user_id,
            connection="businesses",
            count=count,
            limit=limit,
            fields=enf_comma_separated(field="fields", value=fields),
        )
        if return_json:
            return data
        else:
            return BusinessResponse.new_from_json_dict(data)



"""
    Apis for basic media
"""

from typing import Dict, Optional, Union

import pyfacebook.utils.constant as const
from pyfacebook.api.base_resource import BaseResource
from pyfacebook.models.ig_basic_models import IgBasicMedia, IgBasicMediaChildren
from pyfacebook.utils.params_utils import enf_comma_separated


class IGBasicMedia(BaseResource):
    def get_info(
        self,
        media_id: str,
        fields: Optional[Union[str, list, tuple]] = None,
        return_json: bool = False,
    ) -> Union[IgBasicMedia, dict]:
        """
        Represents an image, video, or album media data.

        :param media_id: ID for Media.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a dataclass for IgBasicMedia.
            Or return json data. Default is false.
        :return: Basic media information.
        """

        if fields is None:
            fields = const.IG_BASIC_MEDIA_FIELDS

        data = self.client.get_object(
            object_id=media_id,
            fields=enf_comma_separated(field="fields", value=fields),
        )

        if return_json:
            return data
        else:
            return IgBasicMedia.new_from_json_dict(data)

    def get_batch(
        self,
        ids: Optional[Union[str, list, tuple]],
        fields: Optional[Union[str, list, tuple]] = None,
        return_json: bool = False,
    ) -> Union[Dict[str, IgBasicMedia], dict]:
        """
        Get batch basic media information by ids

        :param ids: IDs for the medias.
        :param fields Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a dataclass for IgBasicMedia.
            Or return json data. Default is false.
        :return: Basic medias information.
        """

        ids = enf_comma_separated(field="ids", value=ids)
        if fields is None:
            fields = const.IG_BUSINESS_MEDIA_PUBLIC_FIELDS
        data = self.client.get_objects(
            ids=ids, fields=enf_comma_separated(field="fields", value=fields)
        )
        if return_json:
            return data
        else:
            return {
                media_id: IgBasicMedia.new_from_json_dict(item)
                for media_id, item in data.items()
            }

    def get_children(
        self,
        media_id: str,
        fields: Optional[Union[str, list, tuple]] = None,
        return_json: bool = False,
    ) -> Union[IgBasicMediaChildren, dict]:
        """
        Get a collection of image and video Media on an album Media.

        :param media_id: ID for the album media.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a dataclass for IgBasicMediaChildren.
            Or return json data. Default is false.
        :return: Media children response information.
        """

        if fields is None:
            fields = const.IG_BASIC_MEDIA_CHILDREN_FIELDS

        data = self.client.get_connection(
            object_id=media_id,
            connection="children",
            fields=enf_comma_separated(field="fields", value=fields),
        )
        if return_json:
            return data
        else:
            return IgBasicMediaChildren.new_from_json_dict(data)



"""
    Apis for business.
"""

from typing import Dict, Optional, Union

import pyfacebook.utils.constant as const
from pyfacebook.api.base_resource import BaseResource
from pyfacebook.models.business import Business
from pyfacebook.utils.params_utils import enf_comma_separated


class FacebookBusiness(BaseResource):
    def get_info(
        self,
        business_id: str,
        fields: Optional[Union[str, list, tuple]] = None,
        return_json: bool = False,
    ) -> Union[Business, dict]:
        """
        Get information about a Facebook business.

        :param business_id: ID for the business.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a dataclass for Business.
            Or return json data. Default is false.
        :return: Business information.
        """
        if fields is None:
            fields = const.BUSINESS_PUBLIC_FIELDS

        data = self.client.get_object(
            object_id=business_id,
            fields=enf_comma_separated(field="fields", value=fields),
        )
        if return_json:
            return data
        else:
            return Business.new_from_json_dict(data=data)

    def get_batch(
        self,
        ids: Union[str, list, tuple],
        fields: Optional[Union[str, list, tuple]] = None,
        return_json: bool = False,
    ) -> Union[Dict[str, Business], dict]:
        """
        Get batch businesses information by ids.

        :param ids: IDs for the business.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a list of dataclass for Businesses.
            Or return json data. Default is false.
        :return: Businesses information.
        """
        ids = enf_comma_separated(field="ids", value=ids)

        if fields is None:
            fields = const.BUSINESS_PUBLIC_FIELDS

        data = self.client.get_objects(
            ids=ids,
            fields=enf_comma_separated(field="fields", value=fields),
        )
        if return_json:
            return data
        else:
            return {
                business_id: Business.new_from_json_dict(item)
                for business_id, item in data.items()
            }



"""
    Tests for basic media api.
"""

import responses


def test_get_info(helpers, api):
    media_id = "18027939643230671"
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.instagram.com/{api.version}/{media_id}",
            json=helpers.load_json(
                "testdata/instagram_basic/apidata/media/media_info.json"
            ),
        )

        media = api.media.get_info(media_id=media_id)
        assert media.id == media_id

        media_json = api.media.get_info(
            media_id=media_id,
            fields="caption,id,media_type,media_url,permalink,timestamp",
            return_json=True,
        )
        assert media_json["id"] == media_id


def test_get_batch(helpers, api):
    ids = ["18027939643230671", "17846368219941692"]
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.instagram.com/{api.version}",
            json=helpers.load_json(
                "testdata/instagram_basic/apidata/media/medias_info.json"
            ),
        )

        medias = api.media.get_batch(ids=ids)
        assert medias[ids[0]].id == ids[0]

        medias_json = api.media.get_batch(
            ids=ids,
            fields="id,media_type,media_url,permalink,timestamp",
            return_json=True,
        )
        assert medias_json[ids[0]]["id"] == ids[0]


def test_get_children(helpers, api):
    media_id = "18027939643230671"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.instagram.com/{api.version}/{media_id}/children",
            json=helpers.load_json(
                "testdata/instagram_basic/apidata/media/media_children.json"
            ),
        )

        children = api.media.get_children(media_id=media_id)
        assert len(children.data) == 2

        children_json = api.media.get_children(
            media_id=media_id,
            fields="id,media_type,media_url",
            return_json=True,
        )
        assert len(children_json["data"]) == 2



"""
    Tests for user api.
"""

import responses


def test_get_info(helpers, api):
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{api.instagram_business_id}",
            json=helpers.load_json("testdata/instagram/apidata/users/user_fields.json"),
        )

        user = api.user.get_info()
        assert user.id == api.instagram_business_id

        user_json = api.user.get_info(
            fields="id,biography,name,username,profile_picture_url,followers_count,media_count",
            return_json=True,
        )
        assert user_json["id"] == api.instagram_business_id


def test_discovery_user(helpers, api):
    username = "facebookfordevelopers"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{api.instagram_business_id}",
            json=helpers.load_json(
                "testdata/instagram/apidata/discovery/medias_p1.json"
            ),
        )

        user = api.user.discovery_user(username=username)
        assert user.business_discovery.id == "17841407673135339"

        user_json = api.user.discovery_user(
            username=username,
            return_json=True,
        )
        assert user_json["business_discovery"]["id"] == "17841407673135339"


def test_discovery_media(helpers, api):
    username = "facebookfordevelopers"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{api.instagram_business_id}",
            json=helpers.load_json(
                "testdata/instagram/apidata/discovery/medias_p1.json"
            ),
        )
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{api.instagram_business_id}",
            json=helpers.load_json(
                "testdata/instagram/apidata/discovery/medias_p2.json"
            ),
        )

        media_p1 = api.user.discovery_user_medias(username=username, limit=2)
        media_p2 = api.user.discovery_user_medias(
            username=username,
            limit=2,
            after=media_p1.business_discovery.media.paging.cursors.after,
        )
        assert len(media_p1.business_discovery.media.data) == 2
        assert len(media_p2.business_discovery.media.data) == 2

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{api.instagram_business_id}",
            json=helpers.load_json(
                "testdata/instagram/apidata/discovery/medias_p2.json"
            ),
        )

        media = api.user.discovery_user_medias(
            username=username, limit=2, before="before", return_json=True
        )
        assert len(media["business_discovery"]["media"]["data"]) == 2


def test_get_content_publishing_limit(helpers, api):
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{api.instagram_business_id}/content_publishing_limit",
            json=helpers.load_json(
                "testdata/instagram/apidata/users/content_publish_limit.json"
            ),
        )

        limit = api.user.get_content_publishing_limit()
        assert limit.data[0].config.quota_total == 25

        limit_json = api.user.get_content_publishing_limit(return_json=True)
        assert limit_json["data"][0]["quota_usage"] == 0


def test_get_user_insights(helpers, api):
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{api.instagram_business_id}/insights",
            json=helpers.load_json(
                "testdata/instagram/apidata/users/user_insights.json"
            ),
        )

        insights = api.user.get_insights(
            metric="impressions,reach,profile_views",
            period="day",
            access_token=api.access_token,
        )
        assert len(insights.data) == 3

        insights_json = api.user.get_insights(
            metric="impressions,reach,profile_views",
            period="day",
            return_json=True,
        )
        assert len(insights_json["data"]) == 3


def test_get_user_media(helpers, api):
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{api.instagram_business_id}/media",
            json=helpers.load_json("testdata/instagram/apidata/users/medias_p1.json"),
        )
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{api.instagram_business_id}/media",
            json=helpers.load_json("testdata/instagram/apidata/users/medias_p2.json"),
        )

        medias = api.user.get_media(count=3, limit=2)
        assert len(medias.data) == 3
        assert medias.data[0].id == "17895731045244887"

        medias_json = api.user.get_media(
            limit=2,
            return_json=True,
        )
        assert len(medias_json["data"]) == 2


def test_get_mentioned_comment(helpers, api):
    cm_id = "17967008497439572"
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{api.instagram_business_id}",
            json=helpers.load_json(
                "testdata/instagram/apidata/users/mentioned_comment.json"
            ),
        )

        comment = api.user.get_mentioned_comment(comment_id=cm_id)
        assert comment.mentioned_comment.id == cm_id

        comment_json = api.user.get_mentioned_comment(
            comment_id=cm_id, return_json=True
        )
        assert comment_json["mentioned_comment"]["id"] == cm_id


def test_get_mentioned_media(helpers, api):
    media_id = "17889114512354744"
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{api.instagram_business_id}",
            json=helpers.load_json(
                "testdata/instagram/apidata/users/mentioned_media.json"
            ),
        )

        media = api.user.get_mentioned_media(media_id=media_id)
        assert media.mentioned_media.id == media_id

        media_json = api.user.get_mentioned_media(media_id=media_id, return_json=True)
        assert media_json["mentioned_media"]["id"] == media_id


def test_get_hashtag_search(helpers, api):
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/ig_hashtag_search",
            json=helpers.load_json(
                "testdata/instagram/apidata/users/hashtag_search.json"
            ),
        )

        hashtag = api.user.get_hashtag_search(q="developers")
        assert hashtag.data[0].id == "17841562426109234"

        hashtag_json = api.user.get_hashtag_search(q="developers", return_json=True)
        assert hashtag_json["data"][0]["id"] == "17841562426109234"


def test_get_recently_searched_hashtags(helpers, api):
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{api.instagram_business_id}/recently_searched_hashtags",
            json=helpers.load_json(
                "testdata/instagram/apidata/users/recently_searched_hashtags.json"
            ),
        )

        hashtags = api.user.get_recently_searched_hashtags()
        assert len(hashtags.data) == 2
        assert hashtags.data[0].id == "17841562426109234"

        hashtags_json = api.user.get_recently_searched_hashtags(return_json=True)
        assert len(hashtags_json["data"]) == 2


def test_get_stories(helpers, api):
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{api.instagram_business_id}/stories",
            json=helpers.load_json("testdata/instagram/apidata/users/stories.json"),
        )

        stories = api.user.get_stories()
        assert len(stories.data) == 2
        assert stories.data[0].id == "17879450117512729"

        stories_json = api.user.get_stories(return_json=True)
        assert len(stories_json["data"]) == 2


def test_get_tagged_media(helpers, api):
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{api.instagram_business_id}/tags",
            json=helpers.load_json(
                "testdata/instagram/apidata/users/tagged_medias.json"
            ),
        )

        medias = api.user.get_tagged_media()
        assert len(medias.data) == 5
        assert medias.data[0].id == "17892354025952911"

        medias_json = api.user.get_tagged_media(return_json=True)
        assert len(medias_json["data"]) == 5


def test_get_live_media(helpers, api):
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{api.instagram_business_id}/live_media",
            json=helpers.load_json("testdata/instagram/apidata/users/live_medias.json"),
        )

        medias = api.user.get_live_media()
        assert len(medias.data) == 1
        assert medias.data[0].id == "90010498116233"

        medias_json = api.user.get_live_media(return_json=True)
        assert len(medias_json["data"]) == 1


def test_get_available_catalogs(helpers, api):
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{api.instagram_business_id}/available_catalogs",
            json=helpers.load_json(
                "testdata/instagram/apidata/users/available_catalogs.json"
            ),
        )

        catalogs = api.user.get_available_catalogs()
        assert len(catalogs.data) == 1
        assert catalogs.data[0].catalog_id == "960179311066902"

        catalogs_json = api.user.get_available_catalogs(return_json=True)
        assert len(catalogs_json["data"]) == 1


def test_get_catalog_product_search(helpers, api):
    catalog_id = "960179311066902"
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{api.instagram_business_id}/catalog_product_search",
            json=helpers.load_json(
                "testdata/instagram/apidata/users/catalog_product_search.json"
            ),
        )

        products = api.user.get_catalog_product_search(catalog_id=catalog_id)
        assert len(products.data) == 1
        assert products.data[0].product_id == 3231775643511089

        products_json = api.user.get_catalog_product_search(
            catalog_id=catalog_id, return_json=True
        )
        assert len(products_json["data"]) == 1


def test_get_product_appeal(helpers, api):
    product_id = 4029274203846188
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{api.instagram_business_id}/product_appeal",
            json=helpers.load_json(
                "testdata/instagram/apidata/users/product_appeal.json"
            ),
        )

        appeals = api.user.get_product_appeal(product_id=product_id)
        assert len(appeals.data) == 1
        assert appeals.data[0].product_id == product_id

        appeals_json = api.user.get_product_appeal(
            product_id=product_id, return_json=True
        )
        assert len(appeals_json["data"]) == 1



"""
    tests for api base
"""


def test_resource(fb_api):
    assert fb_api.user.access_token == "token"
    assert fb_api.user.app_id == "123456"
    assert fb_api.user.app_secret == "xxxxx"
    assert fb_api.user.client



"""
    Apis for video
"""

from typing import Dict, Optional, Union

import pyfacebook.utils.constant as const
from pyfacebook.api.base_resource import BaseResource
from pyfacebook.models.video import Video
from pyfacebook.utils.params_utils import enf_comma_separated


class FacebookVideo(BaseResource):
    def get_info(
        self,
        video_id: Optional[str],
        fields: Optional[Union[str, list, tuple]] = None,
        return_json: bool = False,
    ) -> Union[Video, dict]:
        """
        Get information about a Facebook Video.

        :param video_id: ID for the video.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a dataclass for Video.
            Or return json data. Default is false.
        :return: Video information.
        """
        if fields is None:
            fields = const.VIDEO_PUBLIC_FIELDS
        data = self.client.get_object(
            object_id=video_id,
            fields=enf_comma_separated(field="fields", value=fields),
        )
        if return_json:
            return data
        else:
            return Video.new_from_json_dict(data=data)

    def get_batch(
        self,
        ids: Optional[Union[str, list, tuple]],
        fields: Optional[Union[str, list, tuple]] = None,
        return_json: bool = False,
    ) -> Union[Dict[str, Video], dict]:
        """
        Get batch photos information by ids.

        :param ids: IDs for the photos.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a dataclass for Photo.
            Or return json data. Default is false.
        :return: Photos information.
        """

        ids = enf_comma_separated(field="ids", value=ids)

        if fields is None:
            fields = const.VIDEO_PUBLIC_FIELDS

        data = self.client.get_objects(
            ids=ids,
            fields=enf_comma_separated(field="fields", value=fields),
        )
        if return_json:
            return data
        else:
            return {
                video_id: Video.new_from_json_dict(item)
                for video_id, item in data.items()
            }



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyfacebook',
 'pyfacebook.api',
 'pyfacebook.api.facebook',
 'pyfacebook.api.facebook.common_edges',
 'pyfacebook.api.facebook.resource',
 'pyfacebook.api.instagram_basic',
 'pyfacebook.api.instagram_basic.resource',
 'pyfacebook.api.instagram_business',
 'pyfacebook.api.instagram_business.resource',
 'pyfacebook.models',
 'pyfacebook.utils',
 'tests',
 'tests.facebook',
 'tests.instagram_basic',
 'tests.instagram_business',
 'tests.utils']

package_data = \
{'': ['*']}

install_requires = \
['dataclasses-json>=0.5.4,<0.6.0',
 'requests-oauthlib>=1.3.0,<2.0.0',
 'requests>=2.24.0,<3.0.0']

setup_kwargs = {
    'name': 'python-facebook-api',
    'version': '0.14.4',
    'description': 'A simple Python wrapper around the Facebook Graph API',
    'long_description': 'Python Facebook\n\nA Python wrapper for the Facebook Common API.\n\n.. image:: https://github.com/sns-sdks/python-facebook/workflows/Test/badge.svg\n    :target: https://github.com/sns-sdks/python-facebook/actions\n    :alt: Build Status\n\n.. image:: https://img.shields.io/badge/Docs-passing-brightgreen\n    :target: https://sns-sdks.github.io/python-facebook/\n    :alt: Documentation Status\n\n.. image:: https://codecov.io/gh/sns-sdks/python-facebook/branch/master/graph/badge.svg\n    :target: https://codecov.io/gh/sns-sdks/python-facebook\n    :alt: Codecov\n\n.. image:: https://img.shields.io/pypi/v/python-facebook-api.svg\n    :target: https://pypi.org/project/python-facebook-api\n    :alt: PyPI\n\n\n============\nIntroduction\n============\n\nWe have refactored this library after v0.10.0. If you want to use old version, please see branch ``v0``.\n\nThe new structure we will provide like follow show.\n\n.. image:: docs/docs/images/structure.png\n\n\n.. note::\n\n    This new structure still in developing.\n\n    Now You can use base class ``GraphApi`` to get data.\n\n==========\nInstalling\n==========\n\nIf you want to use old version you can set version to ``0.9.*``, And this series will also support with python2.7\n\nYou can install this library from ``pypi``::\n\n    $pip install --upgrade python-facebook-api\n    ✨🍰✨\n\n\n=====\nUsage\n=====\n\n--------\nGraphAPI\n--------\n\nNow you can use ``GraphApi`` class to communicate with Facebook Graph Api.\n\nYou can initial ``GraphApi`` with three different methods.\n\n1. if you already have an access token, you can initial with it::\n\n    >>> from pyfacebook import GraphAPI\n    >>> api = GraphAPI(access_token="token")\n\n2. if you want to use app credentials to generate app token::\n\n    >>> from pyfacebook import GraphAPI\n    >>> api = GraphAPI(app_id="id", app_secret="secret", application_only_auth=True)\n\n3. if you want to perform an authorization process to a user::\n\n    >>> from pyfacebook import GraphAPI\n    >>> api = GraphAPI(app_id="id", app_secret="secret", oauth_flow=True)\n    >>> api.get_authorization_url()\n    # (\'https://www.facebook.com/dialog/oauth?response_type=code&client_id=id&redirect_uri=https%3A%2F%2Flocalhost%2F&scope=public_profile&state=PyFacebook\', \'PyFacebook\')\n    # let user to do oauth at the browser opened by link.\n    # then get the response url\n    >>> api.exchange_user_access_token(response="url redirected")\n    # Now the api will get the user access token.\n\nThen you can get data from facebook.\n\nGet object data::\n\n    >>> api.get_object(object_id="20531316728")\n    >>> {\'name\': \'Facebook App\', \'id\': \'20531316728\'}\n\nMore you can see the code because we still working on new structure.\n\n-----------\nFacebookAPI\n-----------\n\nInitial methods same with ``GraphAPI``.\n\nGet user data::\n\n    >>> fb.user.get_info(user_id="413140042878187")\n    >>> User(id=\'413140042878187\', name=\'Kun Liu\')\n\nGet page data::\n\n    >>> fb.page.get_info(page_id="20531316728")\n    >>> Page(id=\'20531316728\', name=\'Facebook App\')\n\nSee more in documents.\n\n========\nFeatures\n========\n\n\nNow library has cover follows features\n\nFacebook Graph API:\n\n- Application and Application\'s edges\n- Page and Page\'s edges\n- User and User\'s edges\n- Group and Group\'s edges\n- Event and Event\'s edges\n\nIG Business Graph API:\n\n- User and User\'s edges\n- Media and Media\'s edges\n\nIG Basic Display API:\n\n- User and User\'s edges\n- Media and Media\'s edges\n\n=======\nSUPPORT\n=======\n\n``python-facebook-api`` had been being developed with Pycharm under the free JetBrains Open Source license(s) granted by JetBrains s.r.o.,\nhence I would like to express my thanks here.\n\n.. image:: docs/docs/images/jetbrains.svg\n    :target: https://www.jetbrains.com/?from=sns-sdks/python-facebook\n    :alt: Jetbrains\n',
    'author': 'Ikaros kun',
    'author_email': 'merle.liukun@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/sns-sdks/python-facebook',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""
    Tests for conversation.
"""

import responses


def test_get_info(helpers, fb_api):
    cvs_id = "t_587956915396498"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{cvs_id}",
            json=helpers.load_json(
                "testdata/facebook/apidata/conversations/cvs_info.json"
            ),
        )

        cvs = fb_api.conversation.get_info(conversation_id=cvs_id)
        assert cvs.id == cvs_id
        assert cvs.updated_time == "2021-07-23T09:28:50+0000"

        cvs_json = fb_api.conversation.get_info(
            conversation_id=cvs_id,
            fields="id,link,message_count,snippet,unread_count,updated_time",
            return_json=True,
        )
        assert cvs_json["id"] == cvs_id


def test_get_batch(helpers, fb_api):
    cvs_ids = ["t_587956915396498", "t_233782918546745"]

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}",
            json=helpers.load_json(
                "testdata/facebook/apidata/conversations/cvses_info.json"
            ),
        )

        cvses = fb_api.conversation.get_batch(ids=cvs_ids)
        assert cvses[cvs_ids[0]].id == cvs_ids[0]

        cvses_json = fb_api.conversation.get_batch(
            ids=cvs_ids,
            fields="id,link,message_count,snippet,unread_count,updated_time",
            return_json=True,
        )
        assert cvses_json[cvs_ids[1]]["id"] == cvs_ids[1]



"""
    Model class for Photo.

    Refer: https://developers.facebook.com/docs/graph-api/reference/photo/
"""

from dataclasses import dataclass
from typing import List, Optional

from dataclasses_json import config

from pyfacebook.models.base import BaseModel, field
from pyfacebook.models.extensions import Paging
from pyfacebook.models.attachment import EntityAtTextRange


@dataclass
class PlatformImageSource(BaseModel):
    """
    A class representing the Platform Image Source.

    Refer: https://developers.facebook.com/docs/graph-api/reference/platform-image-source/
    """

    height: Optional[int] = field()
    source: Optional[str] = field(repr=True)
    width: Optional[int] = field()


@dataclass
class Photo(BaseModel):
    """
    A class representing the Photo.
    """

    id: Optional[str] = field(repr=True, compare=True)
    album: Optional[
        dict
    ] = (
        field()
    )  # TODO Refer: https://developers.facebook.com/docs/graph-api/reference/album/
    alt_text: Optional[str] = field()
    alt_text_custom: Optional[str] = field()
    backdated_time: Optional[str] = field()
    backdated_time_granularity: Optional[str] = field()
    can_backdate: Optional[bool] = field()
    can_delete: Optional[bool] = field()
    can_tag: Optional[bool] = field()
    created_time: Optional[str] = field()
    event: Optional[
        dict
    ] = (
        field()
    )  # TODO Refer: https://developers.facebook.com/docs/graph-api/reference/event/
    _from: Optional[dict] = field(metadata=config(field_name="from"))
    height: Optional[int] = field()
    icon: Optional[str] = field()
    images: Optional[List[PlatformImageSource]] = field()
    link: Optional[str] = field()
    name: Optional[str] = field(repr=True)
    name_tags: Optional[List[EntityAtTextRange]] = field()
    page_story_id: Optional[str] = field()
    place: Optional[
        dict
    ] = (
        field()
    )  # TODO Refer: https://developers.facebook.com/docs/graph-api/reference/place/
    target: Optional[
        dict
    ] = (
        field()
    )  # TODO Refer: https://developers.facebook.com/docs/graph-api/reference/profile/
    updated_time: Optional[str] = field()
    webp_images: Optional[List[PlatformImageSource]] = field()
    width: Optional[int] = field()


@dataclass
class PhotosResponse(BaseModel):
    """
    A class representing the result for photos edge.

    Refer: https://developers.facebook.com/docs/graph-api/reference/page/photos
    """

    data: List[Photo] = field(repr=True, compare=True)
    paging: Optional[Paging] = field()



import pytest

from pyfacebook import IGBusinessApi


@pytest.fixture
def api():
    return IGBusinessApi(
        app_id="123456",
        app_secret="xxxxx",
        access_token="token",
        instagram_business_id="17841407673135339",
    )



from pyfacebook.models import *
from pyfacebook.ratelimit import RateLimitHeader, RateLimit, PercentSecond
from pyfacebook.exceptions import PyFacebookException, FacebookError, LibraryError
from pyfacebook.api import GraphAPI, BasicDisplayAPI
from pyfacebook.api.facebook.client import FacebookApi
from pyfacebook.api.instagram_business.client import IGBusinessApi
from pyfacebook.api.instagram_basic.client import IGBasicDisplayApi

__version__ = "0.14.4"



"""
    tests for params utils
"""

import pytest

from pyfacebook import LibraryError
from pyfacebook.utils.params_utils import enf_comma_separated


def test_enf_comma_separated():
    assert enf_comma_separated("fields", "f1") == "f1"
    assert enf_comma_separated("fields", ["f1", "f2"]) == "f1,f2"
    assert enf_comma_separated("fields", ["f1", "f2", "f2"]) == "f1,f2"
    assert enf_comma_separated("fields", ("f1", "f2")) == "f1,f2"

    with pytest.raises(LibraryError):
        enf_comma_separated("id", 1)

    with pytest.raises(LibraryError):
        enf_comma_separated("id", [None, None])



"""
    Model classes for instagram basic display.
"""

from dataclasses import dataclass
from typing import List, Optional

from pyfacebook.models.base import BaseModel, field
from pyfacebook.models.extensions import Paging


@dataclass
class IgBasicUser(BaseModel):
    """
    A class representing the Basic display user.

    Refer: https://developers.facebook.com/docs/instagram-basic-display-api/reference/user#fields
    """

    id: Optional[str] = field(repr=True, compare=True)
    account_type: Optional[str] = field(repr=True)
    media_count: Optional[int] = field()
    username: Optional[str] = field(repr=True)


@dataclass
class IgBasicMediaChild(BaseModel):
    """
    A class representing the Basic display media child.

    Refer: https://developers.facebook.com/docs/instagram-basic-display-api/reference/media#fields
    """

    id: Optional[str] = field(repr=True, compare=True)
    caption: Optional[str] = field()
    media_type: Optional[str] = field()
    media_url: Optional[str] = field(repr=True)
    permalink: Optional[str] = field()
    thumbnail_url: Optional[str] = field()
    timestamp: Optional[str] = field()


@dataclass
class IgBasicMediaChildren(BaseModel):
    """
    A class representing the Basic display media children data.

    Refer: https://developers.facebook.com/docs/instagram-basic-display-api/reference/media/children
    """

    data: Optional[List[IgBasicMediaChild]] = field(repr=True)
    paging: Optional[Paging] = field()


@dataclass
class IgBasicMedia(IgBasicMediaChild):
    """
    A class representing the Basic display media.

    Refer: https://developers.facebook.com/docs/instagram-basic-display-api/reference/media#fields
    """

    username: Optional[str] = field()


@dataclass
class IgBasicMediaResponse(IgBasicMediaChild):
    """
    A class representing the Basic display medias response.

    Refer: https://developers.facebook.com/docs/instagram-basic-display-api/reference/user/media
    """

    data: Optional[List[IgBasicMedia]] = field(repr=True, compare=True)
    paging: Optional[Paging] = field(repr=True)



import json
import logging
from collections import defaultdict
from dataclasses import dataclass
from json.decoder import JSONDecodeError
from typing import List, Optional

from requests.models import CaseInsensitiveDict

logger = logging.getLogger(__name__)


@dataclass
class PercentSecond(object):
    """
    A class representing for percent and sleep seconds mapping.
    If less percent will sleep seconds.
    """

    percent: int
    seconds: int


@dataclass
class RateLimitHeader(object):
    """
    A class representing the rate limit header.
    Refer: https://developers.facebook.com/docs/graph-api/overview/rate-limiting#headers
    """

    call_count: int = 0
    total_cputime: int = 0
    total_time: int = 0
    call_volume: int = 0  # IG basic display has return this, but have no docs for this.
    cpu_time: int = 0  # IG basic display has return this, but have no docs for this.
    type: Optional[str] = None  # only for Business Use Case Rate Limits
    estimated_time_to_regain_access: Optional[
        str
    ] = None  # only for Business Use Case Rate Limits

    def max_percent(self):
        return max(
            self.call_count,
            self.total_cputime,
            self.total_time,
            self.call_volume,
            self.cpu_time,
        )


class RateLimit(object):
    """
    A class representing the rate limit.
    Refer: https://developers.facebook.com/docs/graph-api/overview/rate-limiting
    """

    DEFAULT_TIME_WINDOW = 60 * 60

    def __init__(self):
        """
        Instantiates the RateLimitObject. Takes a json dict as
        kwargs and maps to the object's dictionary. So for something like:

        {
            "app": {
                "call_count": 10,
                "total_cputime": 5,
                "total_time": 4,
            },
            "business: {
                "business-object-id": {
                    "rate-limit-type": {
                        "call_count": 10,
                        "total_cputime": 5,
                        "total_time": 4,
                        "type": "pages",
                        "estimated_time_to_regain_access": 0
                    }
                }
            }
        }

        The RateLimit object will have an attribute 'resources' from which you
        can perform a lookup like:
            api.rate_limit.get_limit()
        or:
            api.rate_limit.get_limit(object_id="123456", endpoint="pages")
        and a RateLimitHeader instance will be returned.
        """
        self.resources = {
            "app": RateLimitHeader(),
            "business": defaultdict(lambda: defaultdict(RateLimitHeader)),
        }

    @staticmethod
    def parse_headers(headers: CaseInsensitiveDict, key: str) -> Optional[dict]:
        """
        Get rate limit information from header for key.
        :param headers: Response headers
        :param key: rate limit key
        :return:
        """
        usage = headers.get(key)
        if usage:
            try:
                data = json.loads(usage)
                return data
            except JSONDecodeError as ex:
                logger.error(
                    f"Exception in parse {key} data error. Usage is: {usage}. errors: {ex}"
                )
                return None
        return None

    def set_limit(self, headers: CaseInsensitiveDict):
        """
        Get rate limit data from response headers. And update to instance.
        :param headers: Response headers
        :return: None
        """
        app_usage = self.parse_headers(headers, "x-app-usage")
        if app_usage is not None:
            self.resources["app"] = RateLimitHeader(**app_usage)

        business_usage = self.parse_headers(headers, "x-business-use-case-usage")
        if business_usage is not None:
            for business_id, items in business_usage.items():
                for item in items:
                    self.resources["business"][business_id][
                        item["type"]
                    ] = RateLimitHeader(**item)

    def get_limit(
        self, object_id: Optional[str] = None, rate_limit_type: Optional[str] = None
    ) -> RateLimitHeader:
        """
        Get a business user case rate limit for object with type or get app rate limit data.

        :param object_id: business object id
        :param rate_limit_type: rate limit type, like pages,instagram,ads_insights...
            All type at https://developers.facebook.com/docs/graph-api/overview/rate-limiting#headers-2
        :return: RateLimitHeader object containing rate limit information.
        """
        if all([object_id, rate_limit_type]):
            return self.resources["business"][object_id][rate_limit_type]

        return self.resources["app"]

    def get_max_percent(self) -> int:
        # TODO Now only check app usage.
        app_usage = self.resources["app"]
        percent = app_usage.max_percent()
        return percent

    def get_sleep_seconds(
        self, sleep_data: Optional[List[PercentSecond]] = None
    ) -> int:
        """
        Get seconds to sleep in requests.

        :param sleep_data:
            the dict for case percent to sleep seconds. ex:
            [
                PercentSecond(percent=20,seconds=2),
                PercentSecond(percent=50,seconds=3),
                PercentSecond(percent=90,seconds=10),
                PercentSecond(percent=100,seconds=1800),
            ]
        :return: sleep seconds
        """
        if isinstance(sleep_data, list):
            max_percent = self.get_max_percent()
            for ps in sleep_data:
                if max_percent <= ps.percent:
                    return ps.seconds
            return 60 * 10  # sleep 10 minutes
        # Default sleep seconds is 0
        return 0



"""
    Models for live video.

    Refer: https://developers.facebook.com/docs/graph-api/reference/live-video/
"""
from dataclasses import dataclass
from typing import List, Optional

from dataclasses_json import config

from pyfacebook.models.base import BaseModel, field
from pyfacebook.models.extensions import Paging


@dataclass
class VideoCopyrightRule(BaseModel):
    """
    A class representing Video Copyright Rule.

    Refer: https://developers.facebook.com/docs/graph-api/reference/video-copyright-rule/
    """

    id: Optional[str] = field(repr=True, compare=True)
    condition_groups: Optional[List[dict]] = field()  # TODO
    copyrights: Optional[List[str]] = field()
    created_date: Optional[str] = field()
    creator: Optional[
        dict
    ] = (
        field()
    )  # TODO Refer: https://developers.facebook.com/docs/graph-api/reference/user/
    is_in_migration: Optional[bool] = field()
    name: Optional[str] = field(repr=True)


@dataclass
class VideoCopyright(BaseModel):
    """
    A class representing the Video Copyright.

    Refer: https://developers.facebook.com/docs/graph-api/reference/video-copyright/
    """

    id: Optional[str] = field(repr=True, compare=True)
    content_category: Optional[str] = field()
    copyright_content_id: Optional[str] = field()
    creator: Optional[
        dict
    ] = (
        field()
    )  # TODO Refer: https://developers.facebook.com/docs/graph-api/reference/user/
    excluded_ownership_segments: Optional[List[dict]] = field()  # TODO
    in_conflict: Optional[bool] = field()
    monitoring_status: Optional[str] = field()
    monitoring_type: Optional[str] = field()
    ownership_countries: Optional[dict] = field()
    reference_file: Optional[dict] = field()
    reference_file_disabled: Optional[bool] = field()
    reference_file_disabled_by_ops: Optional[bool] = field()
    reference_owner_id: Optional[str] = field()
    rule_ids: Optional[List[VideoCopyrightRule]] = field()
    tags: Optional[List[str]] = field()
    whitelisted_ids: Optional[List[str]] = field()


@dataclass
class LiveEncoder(BaseModel):
    """
    A class representing the Live Encoder.

    Refer: https://developers.facebook.com/docs/graph-api/reference/live-encoder/
    """

    id: Optional[str] = field(repr=True, compare=True)
    brand: Optional[str] = field()
    creation_time: Optional[str] = field()
    current_broadcast: Optional["LiveVideo"] = field()
    current_input_stream: Optional["LiveVideoInputStream"] = field()
    device_id: Optional[str] = field()
    last_heartbeat_time: Optional[str] = field()
    model: Optional[str] = field()
    name: Optional[str] = field()
    status: Optional[str] = field()
    version: Optional[str] = field()


@dataclass
class LiveVideoInputStream(BaseModel):
    """
    A class representing the Live Video Input Stream.

    Refer: https://developers.facebook.com/docs/graph-api/reference/live-video-input-stream/
    """

    id: Optional[str] = field(repr=True, compare=True)
    dash_ingest_url: Optional[str] = field()
    dash_preview_url: Optional[str] = field()
    is_master: Optional[bool] = field()
    live_encoder: Optional[LiveEncoder] = field()
    secure_stream_url: Optional[str] = field()
    stream_health: Optional[dict] = field()  # TODO
    stream_id: Optional[str] = field()
    stream_url: Optional[str] = field()


@dataclass
class LiveVideo(BaseModel):
    """
    A class representing the Live Video.
    """

    id: Optional[str] = field(repr=True, compare=True)
    ad_break_config: Optional[dict] = field()
    ad_break_failure_reason: Optional[str] = field()
    broadcast_start_time: Optional[str] = field()
    copyright: Optional[VideoCopyright] = field()
    creation_time: Optional[str] = field()
    dash_ingest_url: Optional[str] = field()
    dash_preview_url: Optional[str] = field()
    description: Optional[str] = field()
    embed_html: Optional[str] = field()
    _from: Optional[dict] = field(metadata=config(field_name="from"))
    ingest_streams: Optional[List[LiveVideoInputStream]] = field()
    is_manual_mode: Optional[bool] = field()
    is_reference_only: Optional[bool] = field()
    live_encoders: Optional[List[LiveEncoder]] = field()
    live_views: Optional[int] = field()
    overlay_url: Optional[str] = field()
    permalink_url: Optional[str] = field()
    planned_start_time: Optional[str] = field()
    seconds_left: Optional[int] = field()
    secure_stream_url: Optional[str] = field()
    status: Optional[str] = field()
    stream_url: Optional[str] = field()
    targeting: Optional[dict] = field()  # TODO
    title: Optional[str] = field(repr=True)
    video: Optional[
        dict
    ] = (
        field()
    )  # TODO Refer: https://developers.facebook.com/docs/graph-api/reference/video/


@dataclass
class LiveVideosResponse(BaseModel):
    """
    A class representing the result for live videos edge.

    Refer: https://developers.facebook.com/docs/graph-api/reference/page/live_videos
    """

    data: List[LiveVideo] = field(repr=True, compare=True)
    paging: Optional[Paging] = field()



"""
    Tests for basic user api.
"""

import responses


def test_get_info(helpers, api):
    uid = "17841406338772941"
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.instagram.com/{api.version}/me",
            json=helpers.load_json(
                "testdata/instagram_basic/apidata/user/user_info.json"
            ),
        )

        user = api.user.get_info()
        assert user.id == uid

        user_json = api.user.get_info(
            fields="account_type,id,media_count,username",
            return_json=True,
        )
        assert user_json["id"] == uid


def test_user_media(helpers, api):
    uid = "17841406338772941"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.instagram.com/{api.version}/{uid}/media",
            json=helpers.load_json(
                "testdata/instagram_basic/apidata/user/user_medias_p1.json"
            ),
        )
        m.add(
            method=responses.GET,
            url=f"https://graph.instagram.com/{api.version}/{uid}/media",
            json=helpers.load_json(
                "testdata/instagram_basic/apidata/user/user_medias_p2.json"
            ),
        )

        media = api.user.get_media(
            user_id=uid,
            count=None,
            limit=2,
        )
        assert len(media.data) == 4
        assert media.data[0].id == "17846368219941692"

        media_json = api.user.get_media(user_id=uid, limit=2, return_json=True)
        assert len(media_json["data"]) == 2
        assert media_json["data"][0]["id"] == "17896189813249754"



"""
    models tests
"""

import pyfacebook.models as md
from pyfacebook.models.base import dict_minus_none_values


def test_base_model():
    bm = md.BaseModel.new_from_json_dict(None)
    assert bm is None

    assert dict_minus_none_values(None) is None


def test_user(helpers):
    user_data = helpers.load_json("testdata/facebook/models/user.json")

    user = md.User.new_from_json_dict(user_data)
    assert user.id == "4"
    assert user.picture.data.width == 50

    assert user.to_dict()


def test_page(helpers):
    page_data = helpers.load_json("testdata/facebook/models/page.json")

    page = md.Page.new_from_json_dict(page_data)
    assert page.id == "20531316728"
    assert page.picture.data.width == 50
    assert page.cover.id == "10159027219496729"
    assert page.location.country == "America"

    assert page.to_dict()


def test_post(helpers):
    post_data = helpers.load_json("testdata/facebook/models/post.json")

    post = md.Post.new_from_json_dict(post_data)
    assert post.id == "565225540184937_4018908568149933"
    assert post.reactions.summary.total_count == 8878
    assert "message" in repr(post)


def test_group(helpers):
    gp_data = helpers.load_json("testdata/facebook/models/group.json")

    group = md.Group.new_from_json_dict(gp_data)
    assert group.id == "2260975870792283"
    assert group.member_count == 744
    assert group.cover.id == "10156542097347597"


def test_event(helpers):
    event_data = helpers.load_json("testdata/facebook/models/event.json")

    event = md.Event.new_from_json_dict(event_data)
    assert event.id == "5971414212932788"
    assert event.cover.id == "2966383383576806"


def test_photo(helpers):
    ph_data = helpers.load_json("testdata/facebook/models/photo.json")

    photo = md.Photo.new_from_json_dict(ph_data)
    assert photo.id == "10158249017468553"
    assert photo.images[0].width == 1369
    assert photo._from["id"] == "19292868552"


def test_album(helpers):
    ab_data = helpers.load_json("testdata/facebook/models/album.json")

    album = md.Album.new_from_json_dict(ab_data)
    assert album.id == "10153867132423553"
    assert album._from["id"] == "19292868552"


def test_video(helpers):
    v_data = helpers.load_json("testdata/facebook/models/video.json")

    video = md.Video.new_from_json_dict(v_data)
    assert video.id == "1192957457884299"
    assert video.privacy.description == "Public"
    assert video._from["id"] == "19292868552"
    assert video.length == 12.891


def test_live_video(helpers):
    v_data = helpers.load_json("testdata/facebook/models/live_video.json")

    live_video = md.LiveVideo.new_from_json_dict(v_data)
    assert live_video.id == "2943409379207540"
    assert live_video.ingest_streams[0].stream_id == "0"


def test_comment(helpers):
    c_data = helpers.load_json("testdata/facebook/models/comment.json")

    comment = md.Comment.new_from_json_dict(c_data)
    assert comment.id == "2954961744718970_2966313500250461"
    assert comment.attachment.type == "photo"


def test_conversation(helpers):
    cvs_data = helpers.load_json("testdata/facebook/models/conversation.json")

    cvs = md.Conversation.new_from_json_dict(cvs_data)
    assert cvs.id == "t_587956915396498"
    assert cvs.participants.data[0].id == "2711636948859886"


def test_message(helpers):
    msg_data = helpers.load_json("testdata/facebook/models/message.json")

    message = md.Message.new_from_json_dict(msg_data)
    assert (
        message.id
        == "m_ToF35NI1OImBjyUIgplSaBMylUFmkYY4bHogy9C1otLISU6SGhccB5NK-THX_W4EdVQiKVv5SgCW9m-_C78mRA"
    )
    assert message.attachments.data[0].id == "651143212510248"


def test_business(helpers):
    business_data = helpers.load_json("testdata/facebook/models/business.json")

    business = md.Business.new_from_json_dict(business_data)
    assert business.id == "987654321"
    assert business.created_by["id"] == "123456789"


def test_ig_bus_user(helpers):
    data = helpers.load_json("testdata/instagram/models/ig_user.json")

    user = md.IgBusUser.new_from_json_dict(data)
    assert user.id == "17841407673135339"
    assert user.follows_count == 16


def test_ig_bus_media(helpers):
    data = helpers.load_json("testdata/instagram/models/ig_media.json")

    media = md.IgBusMedia.new_from_json_dict(data)
    assert media.id == "17896129349106152"
    assert media.comments_count == 0
    assert media.owner.id == "17841406338772941"
    assert len(media.children.data) == 3
    assert media.children.data[0].timestamp == "2021-07-26T10:45:19+0000"


def test_ig_bus_comment(helpers):
    data = helpers.load_json("testdata/instagram/models/ig_comment.json")

    comment = md.IgBusComment.new_from_json_dict(data)
    assert comment.id == "17892250648466172"
    assert comment.media.id == "17846368219941692"
    assert len(comment.replies.data) == 1
    assert comment.replies.data[0].id == "17845747489952795"


def test_ig_bus_hashtag(helpers):
    data = helpers.load_json("testdata/instagram/models/ig_hashtag.json")

    hashtag = md.IgBusHashtag.new_from_json_dict(data)
    assert hashtag.id == "17841593698074073"
    assert hashtag.name == "coke"


def test_ig_bus_container(helpers):
    data = helpers.load_json("testdata/instagram/models/ig_container.json")

    container = md.IgBusContainer.new_from_json_dict(data)
    assert container.id == "17889615691921648"
    assert container.status_code == "FINISHED"


def test_ig_bus_publish_limit(helpers):
    data = helpers.load_json("testdata/instagram/models/ig_publish_limit.json")

    limit = md.IgBusPublishLimit.new_from_json_dict(data)
    assert limit.config.quota_total == 25


def test_ig_bus_insight(helpers):
    data = helpers.load_json("testdata/instagram/models/ig_insight.json")

    insight = md.IgBusInsight.new_from_json_dict(data)
    assert insight.name == "impressions"
    assert insight.values[0].value == 32
